// DO-NOT-REMOVE begin-copyright-block 
//
// Redistributions of any form whatsoever must retain and/or include the
// following acknowledgment, notices and disclaimer:
//
// This product includes software developed by Carnegie Mellon University.
//
// Copyright 2010 by Mohammad Alisafaee, Eric Chung, Michael Ferdman, Brian Gold,
// Jangwoo Kim, Pejman Lotfi-Kamran, Onur Kocberber, Ippokratis Pandis, 
// Djordje Jevdjic, Jared Smolens, Stephen Somogyi, Evangelos Vlachos, 
// Stavros Volos, Jason Zebchuk, Babak Falsafi, Nikos Hardavellas and Tom Wenisch 
// for the SimFlex Project, Computer Architecture Lab at Carnegie Mellon,
// Carnegie Mellon University.
//
// For more information, see the SimFlex project website at:
//   http://www.ece.cmu.edu/~simflex
//
// You may not use the name "Carnegie Mellon University" or derivations
// thereof to endorse or promote products derived from this software.
//
// If you modify the software you must place a notice on or within any
// modified version provided or made available to any third party stating
// that you have modified the software.  The notice shall include at least
// your name, address, phone number, email address and the date and purpose
// of the modification.
//
// THE SOFTWARE IS PROVIDED "AS-IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER
// EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT LIMITED TO ANY WARRANTY
// THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS OR BE ERROR-FREE AND ANY
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
// TITLE, OR NON-INFRINGEMENT.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
// BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO DIRECT, INDIRECT,
// SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM, OR IN
// ANY WAY CONNECTED WITH THIS SOFTWARE (WHETHER OR NOT BASED UPON WARRANTY,
// CONTRACT, TORT OR OTHERWISE).
//
// DO-NOT-REMOVE end-copyright-block   

#include <boost/dynamic_bitset.hpp>
#include <core/flexus.hpp>

#include <core/simulator_layout.hpp>
#include <core/performance/profile.hpp>
#include <core/simics/configuration_api.hpp>

#include <core/boost_extensions/intrusive_ptr.hpp>

#include <core/target.hpp>
#include <core/debug/debug.hpp>
#include <core/types.hpp>
#include <core/stats.hpp>

#include <components/Common/Transports/MemoryTransport.hpp>
#include <components/Common/MessageQueues.hpp>

#include <components/CMPCache/CMPCacheController.hpp>

#include <fstream>

#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/gzip.hpp>

namespace nCMPCache {

	#define	MAX(a,b) ( ((a) > (b)) ? (a) : (b) ) 

	CMPCacheController::CMPCacheController(	const std::string &aName,
												int32_t aNumCores,
												int32_t aBlockSize,
												int32_t aBankNum,
												int32_t aNumBanks,
												int32_t aBankInterleaving,
												int32_t aNumGroups,
												int32_t aGroupInterleaving,
												int32_t aDirLatency,
												int32_t aDirIssueLat,
												int32_t aTagLatency,
												int32_t aTagIssueLat,
												int32_t aDataLatency,
												int32_t aDataIssueLat,
												bool anEvictClean,
												int32_t aQueueSize,
												int32_t aMAFSize,
												int32_t aDirEBSize,
												int32_t aCacheEBSize,
												std::string &aPolicy,
												std::string &aDirType,
												std::string &aDirConfig,
												std::string &aCacheConfig,
												tFillLevel	aCacheLevel
												)
		: theName(aName)
		, theCMPCacheInfo(aBankNum, aName, aPolicy, aDirType, aDirConfig, aCacheConfig, aNumCores, 
							aBlockSize, aNumBanks, aBankInterleaving, aNumGroups, aGroupInterleaving, 
							aMAFSize, aDirEBSize, aCacheEBSize, anEvictClean, aCacheLevel)
		, theMAFPipeline(aName + "-MafPipeline", 1, 1, 0, 
			boost::intrusive_ptr<Stat::StatLog2Histogram>(new Stat::StatLog2Histogram(aName + "-MafPipeline" + "-InterArrivalTimes")))
		, theDirTagPipeline(aName + "-DirPipeline", 1, MAX(aDirIssueLat,aTagIssueLat), MAX(aDirLatency,aTagLatency), 
			boost::intrusive_ptr<Stat::StatLog2Histogram>(new Stat::StatLog2Histogram(aName + "-DirPipeline" + "-InterArrivalTimes")))
		, theDataPipeline(aName + "-DataPipeline", 1, aDataIssueLat, aDataLatency,
			boost::intrusive_ptr<Stat::StatLog2Histogram>(new Stat::StatLog2Histogram(aName + "-DataPipeline" + "-InterArrivalTimes")))
		, theMafUtilization(aName + "-MafUtilization")
		, theDirTagUtilization(aName + "-DirTagUtilization")
		, theDataUtilization(aName + "-DataUtilization")
		, RequestIn(aQueueSize)
		, SnoopIn(aQueueSize)
		, ReplyIn(aQueueSize)
		, RequestOut(aQueueSize)
		, SnoopOut(aQueueSize)
		, ReplyOut(aQueueSize)
		, theIdleWorkScheduled(false)
	{
DBG_(Dev, ( << "GI = " << aGroupInterleaving << ", " << theCMPCacheInfo.theGroupInterleaving ));
		thePolicy =  AbstractFactory<AbstractPolicy,CMPCacheInfo>::createInstance(aPolicy, theCMPCacheInfo);
		theMaxSnoopsPerRequest = thePolicy->maxSnoopsPerRequest();
	}

	CMPCacheController::~CMPCacheController() {
		delete thePolicy;
	}

	bool CMPCacheController::isQuiesced() const {
		return		theMAFPipeline.empty()
				&&	theDirTagPipeline.empty()
				&&	theDataPipeline.empty()
				&&	RequestIn.empty()
				&&	SnoopIn.empty()
				&&	ReplyIn.empty()
				&&	RequestOut.empty()
				&&	SnoopOut.empty()
				&&	ReplyOut.empty()
				&&	thePolicy->isQuiesced();
	}

	void CMPCacheController::saveState(std::string const & aDirName) {
	}

	void CMPCacheController::loadState(std::string const & aDirName) {
        std::string fname( aDirName);
		// This is a little ugly.
		// The functional simulator saves all the cmp cache/directory state in single files
		// but the timing simulator has seperate objects for each bank
		// This allows the timing simulator to use the same data for multiple different banking schemes

		// Key point, change the name from "XX-L3" to "sys-L3"

		std::string obj_name = theName;
		obj_name.replace(0, theName.find('-'), "sys");

		fname += "/" + obj_name + "-dir.gz";
        std::ifstream ifs(fname.c_str(), std::ios::binary);
        if (! ifs.good()) {
            DBG_( Dev, ( << " saved checkpoint state " << fname << " not found.  Resetting to empty cache. " )  );
        } else {

            boost::iostreams::filtering_stream<boost::iostreams::input> in;
            in.push(boost::iostreams::gzip_decompressor());
            in.push(ifs);

            if ( ! thePolicy->loadDirState( in ) ) {
                DBG_ ( Dev, ( << "Error loading checkpoint state from file: " << fname <<
                           ".  Make sure your checkpoints match your current cache configuration." ) );
                DBG_Assert ( false );
            }
        }
		ifs.close();

		fname = aDirName + "/" + obj_name + "-cache.gz";
        std::ifstream c_ifs(fname.c_str(), std::ios::binary);
        if (! c_ifs.good()) {
            DBG_( Dev, ( << " saved checkpoint state " << fname << " not found.  Resetting to empty cache. " )  );
        } else {

            boost::iostreams::filtering_stream<boost::iostreams::input> in;
            in.push(boost::iostreams::gzip_decompressor());
            in.push(c_ifs);

            if ( ! thePolicy->loadCacheState( in ) ) {
                DBG_ ( Dev, ( << "Error loading checkpoint state from file: " << fname <<
                           ".  Make sure your checkpoints match your current cache configuration." ) );
                DBG_Assert ( false );
            }
        }
		c_ifs.close();
	}

	void CMPCacheController::processMessages() {

		// Track utilization stats
		theMafUtilization << std::make_pair( theMAFPipeline.size(), 1);
		theDirTagUtilization << std::make_pair( theDirTagPipeline.size(), 1);

		// Move messages from the input queues into the MAF pipeline
		// handles priorities and resource reservations
		// messages stall in the queues until all the resources are available
		// and they are the highest priority message that is ready
		scheduleNewProcesses();

		// Take processes out of the MAF pipeline and perform a lookup to determine
		// what action should be taken, and then take that action
		advanceMAFPipeline();

		// take processes out of the directory pipeline
		// and put any output messages in the appropriate queues
		advancePipeline();
	}

	// Examine input queues and move messages into the MAF pipeline when resources are available
	void CMPCacheController::scheduleNewProcesses() {
		FLEXUS_PROFILE();

		bool scheduled = false;

		// First, look for Woken MAF entries
		while (		theMAFPipeline.serverAvail()
				&& thePolicy->MAF().hasWakingEntry()
				&& SnoopOut.hasSpace(theMaxSnoopsPerRequest)
				&& ReplyOut.hasSpace(1)
				&& RequestOut.hasSpace(1)
				&& thePolicy->EBHasSpace(thePolicy->MAF().peekWakingMAF()->transport())
				)
		{
			ProcessEntry_p process(new ProcessEntry( eProcWakeMAF, thePolicy->MAF().getWakingMAF()));
			reserveSnoopOut( process, theMaxSnoopsPerRequest);
			reserveReplyOut( process, 1);
			reserveRequestOut( process, 1);
			reserveEB( process, thePolicy->getEBRequirements(process->transport()) );
			theMAFPipeline.enqueue(process);

			DBG_(Trace, ( << theName << " Scheduled WakeMAF: " << *(process->transport()[MemoryMessageTag]) ));
			scheduled=true;
		}
		if (!scheduled && thePolicy->MAF().hasWakingEntry()) {
			DBG_(Trace, ( << theName << " Failed to schedule waking MAF: SnoopOut " << SnoopOut.getSize() << "/" << SnoopOut.getReserve() 
						  << ", ReplyOut " << ReplyOut.getSize() << "/" << ReplyOut.getReserve() 	
						  << ", ReqOut " << RequestOut.getSize() << "/" << RequestOut.getReserve() 
						  << ", EB full " << std::boolalpha << thePolicy->EBFull() ));
		}

		// We want to make sure we have at least one free evict buffer so we can keep handling incoming requests
		// While the "idleWork" below will free up evict buffers, we also need a higher priority mechanism for
		// Keeping one evict buffer free
		// Do the Cache Evict Buffer and Directory Evict Buffer separately


		while (		theMAFPipeline.serverAvail()
				&& !thePolicy->freeCacheEBPending()	// "Pending" to account for scheduled evicts
				&& (thePolicy->CacheEB().usedEntries() > 0) // Make sure we have used entries available
				&& SnoopOut.hasSpace(1)
				)
		{
			ProcessEntry_p process(new ProcessEntry( eProcCacheEvict, thePolicy->getCacheEvictTransport()));
			reserveSnoopOut( process, 1);

			theMAFPipeline.enqueue(process);
			DBG_(Trace, ( << theName << " Scheduled Cache Evict: " << *(process->transport()[MemoryMessageTag]) ));
			scheduled=true;
		}
		if (!scheduled && !thePolicy->freeCacheEBPending()) {
			DBG_(Trace, ( << theName << " Failed to schedule evict, no snoop buffer space." ));
		}


		while (		theMAFPipeline.serverAvail()
				&& !thePolicy->freeDirEBPending()	// "Pending" to account for scheduled evicts
				&& thePolicy->DirEB().idleWorkReady()
				&& thePolicy->CacheEBHasSpace()
				&& SnoopOut.hasSpace(1)
				)
		{
			ProcessEntry_p process(new ProcessEntry( eProcDirEvict, thePolicy->getDirEvictTransport()));
			reserveSnoopOut( process, 1);
			reserveCacheEB( process, 1 );

			theMAFPipeline.enqueue(process);
			DBG_(Trace, ( << theName << " Scheduled Dir Evict: " << *(process->transport()[MemoryMessageTag]) ));
			scheduled=true;
		}
		if (!scheduled && !thePolicy->freeDirEBPending()) {
			DBG_(Trace, ( << theName << " Failed to schedule evict, no snoop buffer space." ));
		} else if (!scheduled && thePolicy->DirEB().full()) {
				DBG_(Iface, ( << theName << " Failed to schedule Dir Evict when DirEB FULL: Used/Reserved/Size = " << thePolicy->DirEB().used() << "/" << thePolicy->DirEB().reserved() << "/" << thePolicy->DirEB().size() << ", invalidates pending = " << thePolicy->DirEB().invalidates() << ": " << thePolicy->DirEB().listInvalidates() ));
		}

		// Second, look for new Reply Messages (highest priority)
		while (		theMAFPipeline.serverAvail()
				&& !ReplyIn.empty()
				&& SnoopOut.hasSpace(1)
				&& ReplyOut.hasSpace(1)
				)
		{
			ProcessEntry_p process(new ProcessEntry( eProcReply, ReplyIn.dequeue()));
			reserveSnoopOut( process, 1);
			reserveReplyOut( process, 1);
			theMAFPipeline.enqueue(process);

			DBG_(Trace, ( << theName << " Scheduled Reply: " << *(process->transport()[MemoryMessageTag]) ));
			scheduled=true;
		}
		if (!scheduled && !ReplyIn.empty()) {
			DBG_(Trace, ( << theName << " Failed to schedule Reply: SnoopOut " << SnoopOut.getSize() << "/" << SnoopOut.getReserve() 
						  << ", ReplyOut " << ReplyOut.getSize() << "/" << ReplyOut.getReserve() ));
		}

		// Third, look for new Snoop Messages (second highest priority)
		while (		theMAFPipeline.serverAvail()
				&& !SnoopIn.empty()
				&& SnoopOut.hasSpace(2)
				&& ReplyOut.hasSpace(1)
				&& !thePolicy->MAF().full()
				)
		{
			ProcessEntry_p process(new ProcessEntry( eProcSnoop, SnoopIn.dequeue()));
			reserveSnoopOut( process, 2);
			reserveReplyOut( process, 1);
			reserveMAF( process );
			theMAFPipeline.enqueue(process);

			DBG_(Trace, ( << theName << " Scheduled Snoop: " << *(process->transport()[MemoryMessageTag]) ));
			scheduled=true;
		}
		if (!scheduled && !SnoopIn.empty()) {
			DBG_(Trace, ( << theName << " Failed to schedule Snoop: SnoopOut " << SnoopOut.getSize() << "/" << SnoopOut.getReserve() 
						  << ", ReplyOut " << ReplyOut.getSize() << "/" << ReplyOut.getReserve() << std::boolalpha << ", MAF full " << thePolicy->MAF().full() ));
		}

		// Fourth, look for new Request Messages (lowest priority)
		while (		theMAFPipeline.serverAvail()
				&& !RequestIn.empty()
				&& SnoopOut.hasSpace(theMaxSnoopsPerRequest)
				&& ReplyOut.hasSpace(1)
				&& RequestOut.hasSpace(1)
				&& thePolicy->EBHasSpace(RequestIn.peek())
				&& !thePolicy->MAF().full()
				)
		{
			ProcessEntry_p process(new ProcessEntry( eProcRequest, RequestIn.dequeue()));
			reserveSnoopOut( process, theMaxSnoopsPerRequest);
			reserveReplyOut( process, 1);
			reserveRequestOut( process, 1);
			reserveEB( process, thePolicy->getEBRequirements(process->transport()) );
			reserveMAF( process );
			theMAFPipeline.enqueue(process);

			DBG_(Trace, ( << theName << " Scheduled Request: " << *(process->transport()[MemoryMessageTag]) ));
			scheduled=true;
		}
		if (!scheduled && !RequestIn.empty()) {
			DBG_(Iface, ( << theName << " Failed to schedule Request: SnoopOut " << SnoopOut.getSize() << "/" << SnoopOut.getReserve() 
						  << ", ReplyOut " << ReplyOut.getSize() << "/" << ReplyOut.getReserve() 
						  << ", ReqOut " << RequestOut.getSize() << "/" << RequestOut.getReserve() 
						  << ", EB full " << std::boolalpha << thePolicy->EBFull() << ", MAF full " << thePolicy->MAF().full() ));
			if (thePolicy->EBFull()) {
				DBG_(Iface, ( << theName << " DirEB.full() = " << std::boolalpha << thePolicy->DirEB().full() << ", CacheEB (reserved,used,size) = (" << thePolicy->CacheEB().reservedEntries() << ", " << thePolicy->CacheEB().usedEntries() << ", " << theCMPCacheInfo.theCacheEBSize << "), ERB available = " << thePolicy->arrayEvictResourcesAvailable() << ", freeDirEBPending() = " << thePolicy->freeDirEBPending() << ", Dir.idleWorkReady() = " << thePolicy->DirEB().idleWorkReady() ));
			}
		}

		// Fifth, look for idle work (ie. invalidates for maintaining inclusion, etc.)
		while (		theMAFPipeline.serverAvail()
				&& thePolicy->CacheEBHasSpace()
				&& thePolicy->hasIdleWorkAvailable()
				&& SnoopOut.hasSpace(1)
				&& !theIdleWorkScheduled	// Only schedule 1 idle work item at a time
				)
		{
			ProcessEntry_p process(new ProcessEntry( eProcIdleWork, thePolicy->getIdleWorkTransport()));
			reserveSnoopOut( process, 1);
			reserveCacheEB( process, 1);
			theMAFPipeline.enqueue(process);

			// Get any additional, policy specific reservations
			thePolicy->getIdleWorkReservations(process);

			DBG_(Trace, ( << theName << " Scheduled IdleWork: " << *(process->transport()[MemoryMessageTag]) ));
			theIdleWorkScheduled = true;
			scheduled=true;
		}
		if (!scheduled && thePolicy->hasIdleWorkAvailable()) {
			DBG_(Trace, ( << theName << " Failed to schedule idle work, no space in Snoop Out" ));
		}

	}

	void CMPCacheController::advanceMAFPipeline() {
		FLEXUS_PROFILE();

		while ( theMAFPipeline.ready() && theDirTagPipeline.serverAvail()) {
			ProcessEntry_p process = theMAFPipeline.dequeue();
			switch(process->type()) {
			case eProcWakeMAF:	runWakeMAFProcess(process);		break;
			case eProcCacheEvict:	runCacheEvictProcess(process);		break;
			case eProcDirEvict:	runDirEvictProcess(process);		break;
			case eProcReply:	runReplyProcess(process);		break;
			case eProcSnoop:	runSnoopProcess(process);		break;
			case eProcRequest:	runRequestProcess(process);		break;
			case eProcIdleWork:	runIdleWorkProcess(process);	break;
			default:
				DBG_Assert(false, ( << "Invalid process type " << process->type() ));
			}
		}
	}

	void CMPCacheController::advancePipeline() {
		FLEXUS_PROFILE();

		while ( theDirTagPipeline.ready() ) {
			ProcessEntry_p process = theDirTagPipeline.peek();
			if (process->requiresData()) {
				if (theDataPipeline.serverAvail()) {
					if (process->transmitAfterTag()) {
						finalizeProcess(process);
					}
					theDirTagPipeline.dequeue();
					theDataPipeline.enqueue(process);
				} else {
					theDirTagPipeline.stall();
					break;
				}
			} else {
				theDirTagPipeline.dequeue();
				finalizeProcess(process);
			}
		}

		while (theDataPipeline.ready()) {
			ProcessEntry_p process = theDataPipeline.dequeue();
			if (!process->transmitAfterTag()) {
				finalizeProcess(process);
			}
		}

	}

	// Determine action for Request Process and enqueue in directory pipeline
	void CMPCacheController::runRequestProcess(ProcessEntry_p process) {
		FLEXUS_PROFILE();

		DBG_(Trace, ( << theName << " handleRequest: " << *process << " for " << *(process->transport()[MemoryMessageTag]) ));

		thePolicy->handleRequest(process);

	
		switch(process->action()) {
		case eFwdAndWaitAck:
			// Un-reserve 1 snoop, 1 reply, 1 eb and 1 maf
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveReplyOut( process, 1);
			unreserveRequestOut( process, 1);
			unreserveEB( process );
			unreserveMAF( process );
			break;
		case eFwdRequest:
			// Un-reserve 1 snoop, 1 reply, 1 eb and 1 maf
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveReplyOut( process, 1);
			unreserveEB( process );
			unreserveMAF( process );
			break;
		case eNotifyAndWaitAck:
			// Un-reserve 1 eb and 1 maf
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveEB( process );
			unreserveMAF( process );
			unreserveRequestOut( process, 1);
			// We need to send a snoop and a reply
			// Create a MAF entry
			break;
		case eReply:
		case eReplyAndRemoveMAF:
			// Un-reserve 2 snoops, 1 eb and 1 maf
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveRequestOut( process, 1);
			unreserveEB( process );
			unreserveMAF( process );
			break;
		case eStall:
			// Un-reserve 2 snoops, 1 reply, 1 eb and 1 maf
if (process->getSnoopTransports().size() > 0) {
DBG_(Crit, ( << theName << " Process stalled while holding SnoopOut reservations: " << *(process->transport()[MemoryMessageTag]) ));
}
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveReplyOut( process, 1);
			unreserveRequestOut( process, 1);
			unreserveEB( process );
			unreserveMAF( process );
			dumpMAFEntries(process->transport()[MemoryMessageTag]->address());
			break;

// Actions for Evict Messages (now come in on Request channel, still sent out on Snoop Channel (no, I don't know why))
		case eFwdAndWakeEvictMAF:
		case eForward:
			unreserveReplyOut( process, 1);
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveRequestOut( process, 1);
			unreserveEB( process );
			unreserveMAF( process );
			// Send a snoop message
			break;

		case eFwdAndReply:
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveRequestOut( process, 1);
			unreserveEB( process );
			unreserveMAF( process );
			break;

		case eWakeEvictMAF:
		case eNoAction:
			// Un-reserve 2 snoops, 1 request, 1 reply, 1 eb and 1 maf
			unreserveReplyOut( process, 1);
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveRequestOut( process, 1);
			unreserveEB( process );
			unreserveMAF( process );
			break;
// End evict actions

		default:
			DBG_Assert( false, ( << "Unexpected action type for Request process: " << process->action() ));
			break;
		}

		DBG_(Trace, ( << theName << " runRequestProcess: " << *process << " for " << *(process->transport()[MemoryMessageTag]) ));

		// Either put into the directory pipeline, or finalize the process
		if (process->lookupRequired()) {
			theDirTagPipeline.enqueue(process);
		} else {
			finalizeProcess(process);
		}
	}

	// Determine action for Snoop Process and enqueue in directory pipeline
	void CMPCacheController::runSnoopProcess(ProcessEntry_p process) {
		FLEXUS_PROFILE();

		DBG_(Trace, ( << theName << " runSnoopProcess: " << *process->transport()[MemoryMessageTag] ));
		thePolicy->handleSnoop(process);
	
		switch(process->action()) {
		case eFwdAndWakeEvictMAF:
		case eForward:
			// Un-reserve 1 reply
			unreserveSnoopOut( process, 2 - process->getSnoopTransports().size());
			unreserveReplyOut( process, 1);
			unreserveMAF( process );
			// Send a snoop message
			break;
		case eReply:
			// Un-reserve 1 snoop
			unreserveSnoopOut( process, 2);
			unreserveMAF( process );
			// Send a reply message (mimic a reply that would have been sent if the evict hadn't happened)
			break;

		case eWakeEvictMAF:
		case eNoAction:
			// Un-reserve 1 snoop and 1 reply
			unreserveSnoopOut( process, 2);
			unreserveReplyOut( process, 1);
			unreserveMAF( process );
			break;

		default:
			DBG_Assert( false, ( << "Unexpected action type for Snoop process: " << process->action() ));
			break;
		}

		// Either put into the directory pipeline, or finalize the process
		if (process->lookupRequired()) {
			theDirTagPipeline.enqueue(process);
		} else {
			finalizeProcess(process);
		}

	}

	// Determine action for Reply Process and enqueue in directory pipeline
	void CMPCacheController::runReplyProcess(ProcessEntry_p process) {
		FLEXUS_PROFILE();

		DBG_(Trace, ( << theName << " runReplyProcess: " << *process->transport()[MemoryMessageTag] ));
		thePolicy->handleReply(process);

		switch(process->action()) {
		case eForward:
		case eFwdAndWakeEvictMAF:
			unreserveReplyOut( process, 1);
			break;
		case eReply:
		case eReplyAndRemoveMAF:
			// Un-reserve 1 snoop
			unreserveSnoopOut( process, 1);
			unreserveEB(process);	// Make sure we get rid of any EB reservations we might have gotten from the MAF
			// Send a reply message (mimic a reply that would have been sent if the evict hadn't happened)
			break;
		case eRemoveAndWakeMAF:
		case eRemoveMAF:
		case eWakeEvictMAF:
		case eNoAction:
			// Un-reserve 1 snoop
			unreserveSnoopOut( process, 1);
			unreserveReplyOut( process, 1);
			unreserveEB(process);	// Make sure we get rid of any EB reservations we might have gotten from the MAF
			break;
		default:
			DBG_Assert( false, ( << "Unexpected action type for Reply process: " << process->action() ));
			break;
		}
		DBG_(Trace, ( << theName << " runReplyProcess: " << *process << " FOR " << *process->transport()[MemoryMessageTag] ));

		// Either put into the directory pipeline, or finalize the process
		if (process->lookupRequired()) {
			theDirTagPipeline.enqueue(process);
		} else {
			finalizeProcess(process);
		}

	}

	// Determine action for Cache Evict Process 
	void CMPCacheController::runCacheEvictProcess(ProcessEntry_p process) {
		FLEXUS_PROFILE();

		DBG_(Trace, ( << theName << " runEvictProcess: " << *process->transport()[MemoryMessageTag] ));
		thePolicy->handleCacheEvict(process);

		DBG_Assert( process->action() == eForward, ( << "Unexpected action type for Evict process: " << process->action() ));

		// Send out a snoop
		// No pipeline, we don't need the directory pipeline (we already used the MAF pipeline)
		finalizeProcess(process);
		
	}

	// Determine action for Dir Evict Process 
	void CMPCacheController::runDirEvictProcess(ProcessEntry_p process) {
		FLEXUS_PROFILE();

		DBG_(Trace, ( << theName << " runEvictProcess: " << *process->transport()[MemoryMessageTag] ));
		thePolicy->handleDirEvict(process);

		DBG_Assert( process->action() == eForward, ( << "Unexpected action type for Evict process: " << process->action() ));

		// Send out a snoop
		// Evictions come from evict buffer, accessed in parallel with MAF pipeline
		finalizeProcess(process);
		
	}

	// Determine action for IdleWork Process 
	void CMPCacheController::runIdleWorkProcess(ProcessEntry_p process) {
		FLEXUS_PROFILE();

		DBG_(Trace, ( << theName << " runIdleWorkProcess: " << *process->transport()[MemoryMessageTag] ));
		thePolicy->handleIdleWork(process);

		// Remove any remaining EB reservations
		unreserveEB( process );

		if (process->action() == eNoAction) {
			unreserveSnoopOut( process, 1);
		} else {
			DBG_Assert( process->action() == eForward, ( << "Unexpected action type for IdleWork process: " << process->action() ));
		}

		theIdleWorkScheduled = false;

		// Let the IdleWork process use either
		// Either put into the directory pipeline, or finalize the process
		if (process->lookupRequired()) {
			theDirTagPipeline.enqueue(process);
		} else {
			finalizeProcess(process);
		}
		
	}

	// Determine action for WakeMAFProcess and enqueue in directory pipeline
	void CMPCacheController::runWakeMAFProcess(ProcessEntry_p process) {
		FLEXUS_PROFILE();
		thePolicy->handleWakeMAF(process);
	
		switch(process->action()) {
		case eFwdAndWaitAck:
			// Un-reserve 1 snoop, 1 reply, 1 eb 
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveReplyOut( process, 1);
			unreserveRequestOut( process, 1);
			unreserveEB( process );
			// We need to forward a snoop message somewhere
			// Update MAF entry
			// Then enqueue in the Directory Pipeline
			break;
		case eNotifyAndWaitAck:
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveRequestOut( process, 1);
			// Un-reserve 1 eb
			unreserveEB( process );
			// We need to send a snoop and a reply
			// Update MAF entry
			// Enqueue in the Directory Pipeline
			break;
		case eReply:
		case eReplyAndRemoveMAF:
			// Un-reserve 2 snoops, 1 eb
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveRequestOut( process, 1);
			unreserveEB( process );
			// Update MAF entry (will be removed when we leave the pipeline)
			break;
		case eRemoveMAF:
			// Remvoe the MAF now
			thePolicy->MAF().remove(process->maf());
			process->setAction(eNoAction);
			// Fall through!
		case eStall:
		case eNoAction:
			// Un-reserve 2 snoops, 1 reply, 1 eb
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveReplyOut( process, 1);
			unreserveEB( process );
			unreserveRequestOut( process, 1);
			// Update MAF entry that will be woken up later
			break;

// Actions for Evict Messages (now come in on Request channel, still sent out on Snoop Channel (no, I don't know why))
		case eFwdAndWakeEvictMAF:
		case eForward:
			unreserveReplyOut( process, 1);
			unreserveSnoopOut( process, theMaxSnoopsPerRequest - process->getSnoopTransports().size());
			unreserveRequestOut( process, 1);
			unreserveEB( process );
			break;

		default:
			DBG_Assert( false, ( << "Unexpected action type for WakeMAF process: " << *process ));
			break;
		}

		DBG_(Trace, ( << theName << " runWakeMAFProcess: " << *process << " FOR " << *process->transport()[MemoryMessageTag] ));

		// Either put into the directory pipeline, or finalize the process
		if (process->lookupRequired()) {
			theDirTagPipeline.enqueue(process, process->requiredLookups());
		} else {
			finalizeProcess(process);
		}

	}

	// Take final process actions, including removing MAF entry
	// and sending final messages
	void CMPCacheController::finalizeProcess(ProcessEntry_p process) {
		FLEXUS_PROFILE();

		DBG_(Trace, ( << "Finalize process: " << *process ));

		switch(process->action()) {
		case eFwdAndWaitAck:
DBG_Assert(process->getSnoopTransports().size() != 0);
			unreserveSnoopOut( process, process->getSnoopTransports().size());
			sendSnoops(process);
			break;
		case eFwdRequest:
			unreserveRequestOut( process, 1);
			sendRequest(process);
			break;
		case eNotifyAndWaitAck:
DBG_Assert(process->getSnoopTransports().size() != 0);
			unreserveSnoopOut( process, process->getSnoopTransports().size());
			sendSnoops(process);
			unreserveReplyOut(process, 1);
DBG_Assert(process->getReplyTransport()[MemoryMessageTag] != NULL, ( << "No reply msg for " << *(process->transport()[MemoryMessageTag]) ));
			sendReply(process);
			break;
		case eFwdAndReply:
DBG_Assert(process->getSnoopTransports().size() != 0);
			unreserveSnoopOut( process, process->getSnoopTransports().size());
			sendSnoops(process);
			unreserveReplyOut(process, 1);
DBG_Assert(process->getReplyTransport()[MemoryMessageTag] != NULL, ( << "No reply msg for " << *(process->transport()[MemoryMessageTag]) ));
			sendReply(process);
			break;
		case eForwardAndRemoveMAF:
DBG_Assert(process->getSnoopTransports().size() != 0);
			unreserveSnoopOut( process, process->getSnoopTransports().size());
			sendSnoops(process);
			//thePolicy->MAF().remove(process->maf());
			thePolicy->MAF().removeFirst(process->transport()[MemoryMessageTag]->address(), process->transport()[DestinationTag]->requester);
			break;
		case eFwdAndWakeEvictMAF:
DBG_Assert(process->getSnoopTransports().size() != 0);
			unreserveSnoopOut( process, process->getSnoopTransports().size());
			sendSnoops(process);
DBG_(Trace, ( << theName << " Waking MAF Waiting on Evict: " << *(process->maf()->transport()[MemoryMessageTag]) ));
			thePolicy->MAF().wakeAfterEvict(process->maf());
			break;
		case eWakeEvictMAF:
DBG_(Trace, ( << theName << " Waking MAF Waiting on Evict: " << *(process->maf()->transport()[MemoryMessageTag]) ));
			thePolicy->MAF().wakeAfterEvict(process->maf());
			break;
		case eForward:
DBG_Assert(process->getSnoopTransports().size() != 0);
			unreserveSnoopOut( process, process->getSnoopTransports().size());
			sendSnoops(process);
			break;
		case eReply:
			unreserveReplyOut(process, 1);
DBG_Assert(process->getReplyTransport()[MemoryMessageTag] != NULL, ( << "No reply msg for " << *(process->transport()[MemoryMessageTag]) ));
			sendReply(process);
			break;
		case eReplyAndRemoveMAF:
			unreserveReplyOut(process, 1);
DBG_Assert(process->getReplyTransport()[MemoryMessageTag] != NULL, ( << "No reply msg for " << *(process->transport()[MemoryMessageTag]) ));
			sendReply(process);
			thePolicy->MAF().removeFirst(process->transport()[MemoryMessageTag]->address(), process->transport()[DestinationTag]->requester);
			//thePolicy->MAF().remove(process->maf());
			thePolicy->wakeMAFs(process->transport()[MemoryMessageTag]->address());
			break;
		case eRemoveMAF:
			thePolicy->MAF().remove(process->maf());
			break;
		case eRemoveAndWakeMAF:
DBG_Assert(process->transport()[MemoryMessageTag]);
DBG_Assert(process->transport()[DestinationTag]);
			thePolicy->MAF().removeFirst(process->transport()[MemoryMessageTag]->address(), process->transport()[DestinationTag]->requester);
			//thePolicy->MAF().remove(process->maf());
			thePolicy->wakeMAFs(process->transport()[MemoryMessageTag]->address());
			break;
		case eStall:
		case eNoAction:
			break;
		}

DBG_Assert(process->getSnoopTransports().size() == 0);
		DBG_Assert(process->noReservations(), ( << "Process finalized while holding reservations: " << *process ));
	}

	void CMPCacheController::sendSnoops(ProcessEntry_p process) {
		std::list<MemoryTransport>& snoops = process->getSnoopTransports();
DBG_(Trace, ( << theName << " enqueueing " << snoops.size() << " snoop transports." ));
		while(!snoops.empty()) {
DBG_(Trace, ( << theName << " enqueuing Snoop msg: " << *(snoops.front()[MemoryMessageTag]) ));
			SnoopOut.enqueue(snoops.front());
			snoops.pop_front();
		}
	}

	void CMPCacheController::sendReply(ProcessEntry_p process) {
		ReplyOut.enqueue(process->getReplyTransport());
	}

	void CMPCacheController::sendRequest(ProcessEntry_p process) {
		RequestOut.enqueue(process->transport());
	}

	void CMPCacheController::dumpMAFEntries(MemoryAddress addr) {
		MissAddressFile::iterator iter = thePolicy->MAF().find(addr);
		for(; iter != thePolicy->MAF().end(); iter++) {
			DBG_(Trace, ( << theName << " MAF contains: " << *(iter->transport()[MemoryMessageTag]) << " in state " << (int)(iter->state()) ));
		}
	}
}; // namespace nCMPCache

