// DO-NOT-REMOVE begin-copyright-block 
//
// Redistributions of any form whatsoever must retain and/or include the
// following acknowledgment, notices and disclaimer:
//
// This product includes software developed by Carnegie Mellon University.
//
// Copyright 2010 by Mohammad Alisafaee, Eric Chung, Michael Ferdman, Brian Gold,
// Jangwoo Kim, Pejman Lotfi-Kamran, Onur Kocberber, Ippokratis Pandis, 
// Djordje Jevdjic, Jared Smolens, Stephen Somogyi, Evangelos Vlachos, 
// Stavros Volos, Jason Zebchuk, Babak Falsafi, Nikos Hardavellas and Tom Wenisch 
// for the SimFlex Project, Computer Architecture Lab at Carnegie Mellon,
// Carnegie Mellon University.
//
// For more information, see the SimFlex project website at:
//   http://www.ece.cmu.edu/~simflex
//
// You may not use the name "Carnegie Mellon University" or derivations
// thereof to endorse or promote products derived from this software.
//
// If you modify the software you must place a notice on or within any
// modified version provided or made available to any third party stating
// that you have modified the software.  The notice shall include at least
// your name, address, phone number, email address and the date and purpose
// of the modification.
//
// THE SOFTWARE IS PROVIDED "AS-IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER
// EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT LIMITED TO ANY WARRANTY
// THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS OR BE ERROR-FREE AND ANY
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
// TITLE, OR NON-INFRINGEMENT.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
// BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO DIRECT, INDIRECT,
// SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM, OR IN
// ANY WAY CONNECTED WITH THIS SOFTWARE (WHETHER OR NOT BASED UPON WARRANTY,
// CONTRACT, TORT OR OTHERWISE).
//
// DO-NOT-REMOVE end-copyright-block   
#ifndef __CMPCACHE_CMP_CACHE_CONTROLLER_HPP__
#define __CMPCACHE_CMP_CACHE_CONTROLLER_HPP__

#include <list>
#include <components/Common/MessageQueues.hpp>
#include <components/CMPCache/ProcessEntry.hpp>
#include <components/CMPCache/AbstractPolicy.hpp>

namespace nCMPCache {

	using namespace Flexus::SharedTypes;
	using namespace nMessageQueues;

	class CMPCacheController {

	private:

		typedef boost::intrusive_ptr<ProcessEntry> ProcessEntry_p;
		typedef PipelineFifo<ProcessEntry_p> Pipeline;

		std::string theName;

		CMPCacheInfo	theCMPCacheInfo;

		Pipeline theMAFPipeline;
		Pipeline theDirTagPipeline;
		Pipeline theDataPipeline;

		Stat::StatInstanceCounter<int64_t> theMafUtilization;
		Stat::StatInstanceCounter<int64_t> theDirTagUtilization;
		Stat::StatInstanceCounter<int64_t> theDataUtilization;

		AbstractPolicy* thePolicy;

		int32_t theMaxSnoopsPerRequest;

	public:
		CMPCacheController(const std::string &aName,
							int32_t aNumCores,
							int32_t aBlockSize,
							int32_t aBankNum,
							int32_t aNumBanks,
							int32_t aBankInterleaving,
							int32_t aNumGroups,
							int32_t aGroupInterleaving,
							int32_t aDirLatency,
							int32_t aDirIssueLat,
							int32_t aTagLatency,
							int32_t aTagIssueLat,
							int32_t aDataLatency,
							int32_t aDataIssueLat,
							bool anEvictClean,
							int32_t aQueueSize,
							int32_t aMAFSize,
							int32_t aDirEBSize,
							int32_t aCacheEBSize,
							std::string &aPolicy,
							std::string &aDirType,
							std::string &aDirConfig,
							std::string &aCacheConfig,
							tFillLevel	aCacheLevel
							);

		~CMPCacheController();

		// Message queues
		MessageQueue<MemoryTransport> RequestIn;
		MessageQueue<MemoryTransport> SnoopIn;
		MessageQueue<MemoryTransport> ReplyIn;
		
		MessageQueue<MemoryTransport> RequestOut;
		MessageQueue<MemoryTransport> SnoopOut;
		MessageQueue<MemoryTransport> ReplyOut;

		bool theIdleWorkScheduled;


		bool isQuiesced() const;

		void saveState(std::string const & aDirName);

		void loadState(std::string const & aDirName);

		void processMessages();

		inline void reserveSnoopOut( ProcessEntry_p process, uint8_t n) {
DBG_(Trace, ( << theName << "reserveSnoopOut(" << (int)n << "): " << *process ));
			SnoopOut.reserve(n);
			process->snoop_out_reserved += n;
		}

		inline void reserveReplyOut( ProcessEntry_p process, uint8_t n) {
			ReplyOut.reserve(n);
			process->reply_out_reserved += n;
		}

		inline void reserveRequestOut( ProcessEntry_p process, uint8_t n) {
			RequestOut.reserve(n);
			process->req_out_reserved += n;
		}

		inline void reserveCacheEB( ProcessEntry_p process, int32_t n) {
			thePolicy->reserveCacheEB(n);
			process->cache_eb_reserved += n;
		}

		inline void reserveEB( ProcessEntry_p process, int32_t n) {
			thePolicy->reserveEB(n);
			process->dir_eb_reserved += n;
			process->cache_eb_reserved += n;
		}

		inline void reserveMAF( ProcessEntry_p process) {
			thePolicy->MAF().reserve();
			process->maf_reserved = true;
		}

		inline void unreserveSnoopOut( ProcessEntry_p process, uint8_t n) {
DBG_(Trace, ( << theName << "unreserveSnoopOut(" << (int)n << "): " << *process ));
			SnoopOut.unreserve(n);
			process->snoop_out_reserved -= n;
		}

		inline void unreserveReplyOut( ProcessEntry_p process, uint8_t n) {
			ReplyOut.unreserve(n);
			process->reply_out_reserved -= n;
		}

		inline void unreserveRequestOut( ProcessEntry_p process, uint8_t n) {
			RequestOut.unreserve(n);
			process->req_out_reserved -= n;
		}

		inline void unreserveCacheEB( ProcessEntry_p process) {
			thePolicy->unreserveCacheEB(process->cache_eb_reserved);
			process->cache_eb_reserved = 0;
		}

		inline void unreserveEB( ProcessEntry_p process) {
			thePolicy->unreserveDirEB(process->dir_eb_reserved);
			thePolicy->unreserveCacheEB(process->cache_eb_reserved);
			process->dir_eb_reserved = 0;
			process->cache_eb_reserved = 0;
		}

		inline void unreserveMAF( ProcessEntry_p process) {
			thePolicy->MAF().unreserve();
			process->maf_reserved = false;
		}

		void dumpMAFEntries(MemoryAddress addr);

	protected:

		void scheduleNewProcesses();
		void advanceMAFPipeline();
		void advancePipeline();
		void finalizeProcess(ProcessEntry_p process);

		void runRequestProcess(ProcessEntry_p process);
		void runSnoopProcess(ProcessEntry_p process);
		void runReplyProcess(ProcessEntry_p process);
		void runCacheEvictProcess(ProcessEntry_p process);
		void runDirEvictProcess(ProcessEntry_p process);
		void runIdleWorkProcess(ProcessEntry_p process);
		void runWakeMAFProcess(ProcessEntry_p process);

		void sendRequest(ProcessEntry_p process);
		void sendSnoops(ProcessEntry_p process);
		void sendReply(ProcessEntry_p process);

	};	// CMPCacheController

};

#endif // !__CMPCACHE_CMP_CACHE_CONTROLLER_HPP__
