// DO-NOT-REMOVE begin-copyright-block 
//
// Redistributions of any form whatsoever must retain and/or include the
// following acknowledgment, notices and disclaimer:
//
// This product includes software developed by Carnegie Mellon University.
//
// Copyright 2010 by Mohammad Alisafaee, Eric Chung, Michael Ferdman, Brian Gold,
// Jangwoo Kim, Pejman Lotfi-Kamran, Onur Kocberber, Ippokratis Pandis, 
// Djordje Jevdjic, Jared Smolens, Stephen Somogyi, Evangelos Vlachos, 
// Stavros Volos, Jason Zebchuk, Babak Falsafi, Nikos Hardavellas and Tom Wenisch 
// for the SimFlex Project, Computer Architecture Lab at Carnegie Mellon,
// Carnegie Mellon University.
//
// For more information, see the SimFlex project website at:
//   http://www.ece.cmu.edu/~simflex
//
// You may not use the name "Carnegie Mellon University" or derivations
// thereof to endorse or promote products derived from this software.
//
// If you modify the software you must place a notice on or within any
// modified version provided or made available to any third party stating
// that you have modified the software.  The notice shall include at least
// your name, address, phone number, email address and the date and purpose
// of the modification.
//
// THE SOFTWARE IS PROVIDED "AS-IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER
// EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT LIMITED TO ANY WARRANTY
// THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS OR BE ERROR-FREE AND ANY
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
// TITLE, OR NON-INFRINGEMENT.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
// BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO DIRECT, INDIRECT,
// SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM, OR IN
// ANY WAY CONNECTED WITH THIS SOFTWARE (WHETHER OR NOT BASED UPON WARRANTY,
// CONTRACT, TORT OR OTHERWISE).
//
// DO-NOT-REMOVE end-copyright-block   
#ifndef __PROCESS_ENTRY_HPP__
#define __PROCESS_ENTRY_HPP__

#include <list>
#include <core/boost_extensions/intrusive_ptr.hpp>

#include <components/CMPCache/MissAddressFile.hpp>

namespace nCMPCache {

using namespace Flexus::SharedTypes;

typedef Flexus::SharedTypes::PhysicalMemoryAddress MemoryAddress;

// Different types of processes
enum ProcessType {
	eProcWakeMAF,
	eProcCacheEvict,
	eProcDirEvict,
	eProcReply,
	eProcSnoop,
	eProcRequest,
	eProcIdleWork
};

// Each process can carry-out one of these actions
enum ProcessAction {
	eFwdAndWaitAck,
	eNotifyAndWaitAck,
	eForwardAndRemoveMAF,
	eFwdAndWakeEvictMAF,
	eWakeEvictMAF,
	eForward,
	eFwdRequest,
	eFwdAndReply,
	eReply,
	eReplyAndRemoveMAF,
	eRemoveMAF,
	eRemoveAndWakeMAF,
	eStall,
	eNoAction
};

class ProcessEntry : public boost::counted_base {
private:
	ProcessType					theType;
	ProcessAction				theAction;

	MissAddressFile::iterator	theMAFEntry;

	MemoryTransport				theOrigTransport;
	MemoryTransport				theReplyTransport;
	std::list<MemoryTransport>	theSnoopTransports;
	int							theDirLookupsRequired;
	bool						theRequiresData;
	bool						theTransmitAfterTag;

public:
	// Reservations, seperate fields for cleaner code + more flexibility
	uint8_t snoop_out_reserved;
	uint8_t req_out_reserved;
	uint8_t reply_out_reserved;
	bool maf_reserved;
	int32_t cache_eb_reserved;
	int32_t dir_eb_reserved;
	
	bool remove_MAF;

	const ProcessType& type() const { return theType; }
	const ProcessAction& action() const { return theAction; }
	bool lookupRequired() const { return (theDirLookupsRequired > 0); }

	void addLookup() { theDirLookupsRequired++; }
	void completeLookup() { theDirLookupsRequired++; }
	int32_t requiredLookups() const { return theDirLookupsRequired; }

	bool requiresData() const { return theRequiresData; }
	void setRequiresData(bool requires) { theRequiresData = requires; }

	bool transmitAfterTag() const { return theTransmitAfterTag; }
	void setTransmitAfterTag(bool transmit) { theTransmitAfterTag = transmit; }

	bool noReservations() const { 
		return (snoop_out_reserved == 0)
				&& (reply_out_reserved == 0)
				&& (req_out_reserved == 0)
				&& (maf_reserved == false)
				&& (dir_eb_reserved == false)
				&& (cache_eb_reserved == false);
	}

	ProcessEntry(const ProcessType &aType, const MemoryTransport &aTransport) 
			: theType(aType), theAction(eNoAction), theOrigTransport(aTransport), theDirLookupsRequired(1), theRequiresData(false), theTransmitAfterTag(false),
			  snoop_out_reserved(0), req_out_reserved(0), reply_out_reserved(0), maf_reserved(0), cache_eb_reserved(0), dir_eb_reserved(0)
	{
	}

	ProcessEntry(const ProcessType &aType, MissAddressFile::iterator aMaf) 
			: theType(aType), theAction(eNoAction), theMAFEntry(aMaf), theOrigTransport(aMaf->transport()), theDirLookupsRequired(1), theRequiresData(false), theTransmitAfterTag(false),
			  snoop_out_reserved(0), req_out_reserved(0), reply_out_reserved(0), maf_reserved(0), cache_eb_reserved(0), dir_eb_reserved(0)
	{
	}

	void setAction(ProcessAction anAction) {
		theAction = anAction;
	}

	void setLookupsRequired(int32_t aRequired) {
		theDirLookupsRequired = aRequired;
	}


	MissAddressFile::iterator& maf() {
		return theMAFEntry;
	}

	void setMAF(MissAddressFile::iterator& aMAFEntry) {
		theMAFEntry = aMAFEntry;
	}

	MemoryTransport& transport() {
		return theOrigTransport;
	}

	const MemoryTransport& transport() const {
		return theOrigTransport;
	}

	void addSnoopTransport(MemoryTransport &aTransport) {
		DBG_Assert(aTransport[DestinationTag] != NULL, ( << "MemoryMessage: " << (*aTransport[MemoryMessageTag]) ));
		theSnoopTransports.push_back(aTransport);
	}

	void setReplyTransport(MemoryTransport &aTransport) {
		DBG_Assert(aTransport[DestinationTag] != NULL);
		theReplyTransport = aTransport;
	}

	std::list<MemoryTransport>& getSnoopTransports() {
		return theSnoopTransports;
	}

	MemoryTransport& getReplyTransport() {
		return theReplyTransport;
	}

	int32_t getSnoopReservations() const { return snoop_out_reserved; }
	int32_t getReplyReservations() const { return reply_out_reserved; }
	int32_t getRequestReservations() const { return req_out_reserved; }
	bool getMAFReserved() const { return maf_reserved; }
	int32_t getCacheEBReserved() const { return cache_eb_reserved; }
	bool getDirEBReserved() const { return dir_eb_reserved; }
};

typedef boost::intrusive_ptr<ProcessEntry> ProcessEntry_p;

inline std::ostream& operator<<(std::ostream& os, const ProcessEntry& p) {
	switch(p.type()) {
	case eProcWakeMAF:	os << "ProcWakeMAF: "; break;
	case eProcCacheEvict:	os << "ProcCacheEvict: "; break;
	case eProcDirEvict:	os << "ProcDirEvict: "; break;
	case eProcReply:	os << "ProcReply: "; break;
	case eProcSnoop:	os << "ProcSnoop: "; break;
	case eProcRequest:	os << "ProcRequest: "; break;
	case eProcIdleWork: os << "ProcIdleWork: "; break;
	}

	switch(p.action()) {
	case eFwdAndWaitAck:		os << "FwdAndWaitAck, "; break;
	case eNotifyAndWaitAck:		os << "NotifyAndWaitAck, "; break;
	case eForwardAndRemoveMAF:	os << "ForwardAndRemoveMAF, "; break;
	case eFwdAndWakeEvictMAF:	os << "FwdAndWakeEvictMAF, "; break;
	case eWakeEvictMAF:			os << "WakeEvictMAF, "; break;
	case eForward:				os << "Forward, "; break;
	case eFwdRequest:			os << "FwdRequest, "; break;
	case eFwdAndReply:			os << "FwdAndReply, "; break;
	case eReply:				os << "Reply, "; break;
	case eReplyAndRemoveMAF:	os << "ReplyAndRemoveMAF, "; break;
	case eRemoveMAF:			os << "RemoveMAF, "; break;
	case eRemoveAndWakeMAF:		os << "RemoveAndWakeMAF, "; break;
	case eStall:				os << "Stall, "; break;
	case eNoAction:				os << "NoAction, "; break;
	}

	os << "snoop reservations: " << p.getSnoopReservations() << ", ";
	os << "request reservations: " << p.getRequestReservations() << ", ";
	os << "reply reservations: " << p.getReplyReservations() << ", ";
	if (p.getMAFReserved()) {
		os << "MAF reserved, ";
	}
	if (p.getCacheEBReserved()) {
		os << "Cache EB reserved (" << p.getCacheEBReserved() << "), ";
	}
	if (p.getDirEBReserved()) {
		os << "Dir EB reserved, ";
	}
	if (p.transport()[MemoryMessageTag] != NULL) {
		os << *(p.transport()[MemoryMessageTag]);
	}
	return os;
}

#if 0
inline std::ostream& operator<<(std::ostream& os, const ProcessAction& action) {
	switch(action) {
	case eFwdAndWaitAck:		os << "FwdAndWaitAck, "; break;
	case eNotifyAndWaitAck:		os << "NotifyAndWaitAck, "; break;
	case eForwardAndRemoveMAF:	os << "ForwardAndRemoveMAF, "; break;
	case eFwdAndWakeEvictMAF:	os << "FwdAndWakeEvictMAF, "; break;
	case eWakeEvictMAF:			os << "WakeEvictMAF, "; break;
	case eForward:				os << "Forward, "; break;
	case eFwdRequest:			os << "FwdRequest, "; break;
	case eFwdAndReply:			os << "FwdAndReply, "; break;
	case eReply:				os << "Reply, "; break;
	case eReplyAndRemoveMAF:	os << "ReplyAndRemoveMAF, "; break;
	case eRemoveMAF:			os << "RemoveMAF, "; break;
	case eRemoveAndWakeMAF:		os << "RemoveAndWakeMAF, "; break;
	case eStall:				os << "Stall, "; break;
	case eNoAction:				os << "NoAction, "; break;
	}
	return os;
}
#endif

}; // namespace nCMPCache

#endif // __PROCESS_ENTRY_HPP__
