// DO-NOT-REMOVE begin-copyright-block 
//
// Redistributions of any form whatsoever must retain and/or include the
// following acknowledgment, notices and disclaimer:
//
// This product includes software developed by Carnegie Mellon University.
//
// Copyright 2010 by Mohammad Alisafaee, Eric Chung, Michael Ferdman, Brian Gold,
// Jangwoo Kim, Pejman Lotfi-Kamran, Onur Kocberber, Ippokratis Pandis, 
// Djordje Jevdjic, Jared Smolens, Stephen Somogyi, Evangelos Vlachos, 
// Stavros Volos, Jason Zebchuk, Babak Falsafi, Nikos Hardavellas and Tom Wenisch 
// for the SimFlex Project, Computer Architecture Lab at Carnegie Mellon,
// Carnegie Mellon University.
//
// For more information, see the SimFlex project website at:
//   http://www.ece.cmu.edu/~simflex
//
// You may not use the name "Carnegie Mellon University" or derivations
// thereof to endorse or promote products derived from this software.
//
// If you modify the software you must place a notice on or within any
// modified version provided or made available to any third party stating
// that you have modified the software.  The notice shall include at least
// your name, address, phone number, email address and the date and purpose
// of the modification.
//
// THE SOFTWARE IS PROVIDED "AS-IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER
// EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT LIMITED TO ANY WARRANTY
// THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS OR BE ERROR-FREE AND ANY
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
// TITLE, OR NON-INFRINGEMENT.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
// BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO DIRECT, INDIRECT,
// SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM, OR IN
// ANY WAY CONNECTED WITH THIS SOFTWARE (WHETHER OR NOT BASED UPON WARRANTY,
// CONTRACT, TORT OR OTHERWISE).
//
// DO-NOT-REMOVE end-copyright-block   
#ifndef __RT_DIRECTORY_HPP__
#define __RT_DIRECTORY_HPP__

#include <algorithm>
#include <list>
#include <boost/intrusive_ptr.hpp>
#include <boost/function.hpp>
#include <boost/dynamic_bitset.hpp>
#include <boost/multi_index_container.hpp>
#include <boost/multi_index/key_extractors.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index/sequenced_index.hpp>
#include <boost/multi_index/hashed_index.hpp>
#include <boost/multi_index/mem_fun.hpp>

#include <boost/serialization/serialization.hpp>
#include <boost/serialization/tracking.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/export.hpp>

#include <components/Common/Util.hpp>

#include <components/Common/Serializers.hpp>

using nCommonUtil::log_base2;
using nCommonUtil::AddressHash;
using nCommonSerializers::SerializableRTDirEntry;

using namespace boost::multi_index;
using std::list;
using std::string;
using std::pair;

#include <components/CMPCache/AbstractRegionDirectory.hpp>

namespace nCMPCache {

struct by_order {};
struct by_tag {};
struct by_way {};

struct ULLHash {
	const std::size_t operator()(uint64_t x) const {
		return (std::size_t)x;
	}
};

typedef Flexus::SharedTypes::PhysicalMemoryAddress MemoryAddress;

template<typename _State, typename _EState=_State>
class RTDirectory : public AbstractRegionDirectory<_State,_EState> {
public:
private:

	// RegionTracker (RVA) entry structure
	struct RTEntry {
	public:
        MemoryAddress tag() const {	return MemoryAddress(theTag); }
        uint16_t way() const { return theWay; }
        const _State& state(int32_t offset) const { return block_state[offset]; }
        const std::vector<_State>& state() const { return block_state; }
        int32_t owner() const { return theOwner; }

        bool isProtected() { return protect.any(); }

		RTEntry(int32_t blocksPerRegion, int32_t num_sharers) 
				: block_state(blocksPerRegion,_State(num_sharers)), ways(blocksPerRegion,0)
				, protect(blocksPerRegion,false), valid(blocksPerRegion,false), theOwner(-1) {}

		RTEntry(int32_t blocksPerRegion, int32_t num_sharers, uint64_t tag, int32_t way) 
				: theTag(tag), block_state(blocksPerRegion,_State(num_sharers)), ways(blocksPerRegion,0)
				, protect(blocksPerRegion,false), valid(blocksPerRegion,false), theWay(way), theOwner(-1)   {}

		RTEntry(int32_t blocksPerRegion, int32_t num_sharers, uint64_t tag, int32_t way, int32_t owner) 
				: theTag(tag), block_state(blocksPerRegion,_State(num_sharers)), ways(blocksPerRegion,0)
				, protect(blocksPerRegion,false), valid(blocksPerRegion,false), theWay(way), theOwner(owner)   {}

		~RTEntry() {}

		uint64_t theTag;

		std::vector<_State> block_state;
		std::vector<uint16_t> ways;
		boost::dynamic_bitset<uint16_t> protect;
		boost::dynamic_bitset<uint16_t> valid;

		uint16_t theWay;
		uint16_t theOwner;

		friend class RTDirectory<_State,_EState>;
	};

	// Modifier Objects for updating RTEntry objects in multi-index containers

    struct ChangeRTState {
        ChangeRTState(int32_t offset, const _State &state) : offset(offset), state(state) {}
        void operator()(RTEntry &entry) {
            entry.block_state[offset] = state;
        }
    private:
        int32_t offset;
        const _State &state;
    };
    struct AddRTSharer {
        AddRTSharer(int32_t offset, int32_t sharer) : offset(offset), sharer(sharer) {}
        void operator()(RTEntry &entry) {
            entry.block_state[offset].addSharer(sharer);
        }
    private:
        int32_t offset;
        int32_t sharer;
    };
    struct RemoveRTSharer {
        RemoveRTSharer(int32_t offset, int32_t sharer) : offset(offset), sharer(sharer) {}
        void operator()(RTEntry &entry) {
            entry.block_state[offset].removeSharer(sharer);
        }
    private:
        int32_t offset;
        int32_t sharer;
    };
    struct SetRTSharer {
        SetRTSharer(int32_t offset, int32_t sharer) : offset(offset), sharer(sharer) {}
        void operator()(RTEntry &entry) {
            entry.block_state[offset].setSharer(sharer);
        }
    private:
        int32_t offset;
        int32_t sharer;
    };
    struct ChangeRTProtected {
        ChangeRTProtected(int32_t offset, bool val) : offset(offset), val(val) {}
        void operator()(RTEntry &entry) {
			entry.protect[offset] = val;
        }
    private:
        int32_t offset;
		bool val;
    };

    struct ChangeRTValid {
        ChangeRTValid(int32_t offset, bool val) : offset(offset), val(val) {}
        void operator()(RTEntry &entry) {
			entry.valid[offset] = val;
        }
    private:
        int32_t offset;
		bool val;
    };

    struct InvalidateRTState {
        InvalidateRTState(int32_t offset) : offset(offset) {}
        void operator()(RTEntry &entry) {
            // Only keep the cached bit
            entry.block_state[offset].reset();
            entry.protect[offset] = false;
            entry.valid[offset] = false;
        }
    private:
        int32_t offset;
    };

	struct ChangeRegionOwner {
		ChangeRegionOwner(const int32_t &owner) : owner(owner) {}
		void operator()(RTEntry &entry) {
			entry.theOwner = owner;
		}
	private:
		int32_t owner;
	};

	struct CorrectRTWay {
		CorrectRTWay(int32_t offset, uint16_t way) : offset(offset), way(way) {}
		void operator()(RTEntry  &entry) {
			entry.ways[offset] = way;
		}
	private:
		int32_t offset;
		uint16_t way;
	};

    struct AllocateBlock {
        AllocateBlock(int32_t offset, const _State &state, uint16_t way) : offset(offset), state(state), way(way) {}
        void operator()(RTEntry &entry) {
            entry.block_state[offset] = state;
            entry.ways[offset]  = way;
			entry.valid[offset] = true;
        }
    private:
        int32_t offset;
        const _State &state;
        uint16_t way;
    };


	// Structure to mimic traditional tag array (BST like structure)
	struct BlockEntry {
		uint64_t	tag;
		mutable _State		state;
		uint16_t		way;
		bool				protect;
		bool				valid;

		BlockEntry(uint64_t t, const _State &s, uint16_t w, bool v=false)
				: tag(t), state(s), way(w), protect(false), valid(v) {}
	};

	// Modifier Objects to update BlockEntry objects in multi-index containers

    struct ChangeBlockState {
        ChangeBlockState(const _State &state) : state(state) {}
        void operator()(BlockEntry &entry) {
            entry.state = state;
        }
    private:
        const _State& state;
    };
    struct AddBlockSharer {
        AddBlockSharer(int32_t sharer) : sharer(sharer) {}
        void operator()(BlockEntry &entry) {
            entry.state.addSharer(sharer);
        }
    private:
        int32_t sharer;
    };

    struct RemoveBlockSharer {
        RemoveBlockSharer(int32_t sharer) : sharer(sharer) {}
        void operator()(BlockEntry &entry) {
            entry.state.removeSharer(sharer);
        }
    private:
        int32_t sharer;
    };

    struct SetBlockSharer {
        SetBlockSharer(int32_t sharer) : sharer(sharer) {}
        void operator()(BlockEntry &entry) {
            entry.state.setSharer(sharer);
        }
    private:
        int32_t sharer;
    };

    struct ChangeBlockValid {
        ChangeBlockValid(const bool val) : val(val) {}
        void operator()(BlockEntry &entry) {
            entry.protect = val;
        }
    private:
        const bool val;
    };

    struct ChangeBlockProtected {
        ChangeBlockProtected(const bool val) : val(val) {}
        void operator()(BlockEntry &entry) {
            entry.protect = val;
        }
    private:
        const bool val;
    };

	struct ReplaceBlock {
		ReplaceBlock(uint64_t tag, const _State &state) : tag(tag), state(state) {}
		void operator()(BlockEntry &entry) {
			entry.state = state;
			entry.tag = tag;
			entry.valid = true;
		}
	private:
		uint64_t  tag;
		const _State &state;
	};

	typedef multi_index_container
      < RTEntry
      , indexed_by
            < sequenced < tag<by_order> >
            , hashed_unique
                    < tag<by_tag>
                    , member<RTEntry, uint64_t, &RTEntry::theTag>
					, ULLHash
                    >
			, ordered_non_unique
					< tag<by_way>
					, member<RTEntry, uint16_t, &RTEntry::theWay>
					>
            >
	>
    rt_set_t;

	typedef typename rt_set_t::template nth_index<0>::type::iterator	rt_order_iterator;
	typedef typename rt_set_t::template nth_index<1>::type::iterator	rt_iterator;
	typedef typename rt_set_t::template nth_index<2>::type::iterator	rt_way_iterator;

	typedef typename rt_set_t::template nth_index<1>::type				rt_index;

	typedef multi_index_container
      < BlockEntry
      , indexed_by
            < hashed_unique
                    < tag<by_tag>
                    , member<BlockEntry, uint64_t, &BlockEntry::tag>
					, ULLHash
                    >
            , hashed_unique
                    < tag<by_way>
                    , member<BlockEntry, uint16_t, &BlockEntry::way>
                    >
            >
	>
    block_set_t;

	typedef typename block_set_t::template nth_index<0>::type::iterator	tag_iterator;
	typedef typename block_set_t::template nth_index<1>::type::iterator	way_iterator;

	typedef typename block_set_t::template nth_index<1>::type				way_index;

	class RTLookupResult : public AbstractRegionLookupResult<_State> {
	private:
		RTDirectory		*rt_cache;
		rt_set_t		*rt_set;
		block_set_t		*block_set;
		rt_iterator		theRegion;
		way_iterator	block;
		int				offset;
		uint64_t	tagset;
		bool			block_allocated;
		bool			region_allocated;
		_State			orig_state;

		RTLookupResult(int32_t num_sharers) : rt_cache(NULL),block_allocated(false),region_allocated(false),orig_state(_State(num_sharers)) {}
		RTLookupResult(bool rvalid, bool bvalid, RTDirectory<_State,_EState> *rtc, uint64_t addr, int32_t num_sharers) 
			: rt_cache(rtc),tagset(addr),block_allocated(bvalid),region_allocated(rvalid),orig_state(_State(num_sharers)) {}

		friend class RTDirectory<_State,_EState>;

	public:

		virtual ~RTLookupResult() {}

		virtual void changeState(_State new_state, bool make_MRU, bool make_LRU) {
			rt_cache->updateState(*this, new_state, make_MRU, make_LRU);
		}

		virtual const _State& state( void ) const {
			return (block_allocated ? block->state : orig_state);
		}

		virtual void addSharer( int32_t sharer ) {
			rt_cache->addSharer(*this, sharer);
		}
		virtual void removeSharer( int32_t sharer ) {
			rt_cache->removeSharer(*this, sharer);
		}
		virtual void setSharer( int32_t sharer ) {
			rt_cache->setSharer(*this, sharer);
		}
		virtual void setState( const _State &aNewState ) {
			rt_cache->updateState(*this, aNewState, false, false);
		}

		virtual void setProtected( bool val) {
			rt_cache->setProtected(*this, val);
		}

        virtual bool found( void ) { return block_allocated; }

        virtual MemoryAddress blockAddress ( void ) const { return MemoryAddress(tagset); }

		virtual int32_t owner() const { return (region_allocated ? theRegion->owner() : -1); }

        virtual const _State& regionState(int32_t i) const { DBG_Assert(region_allocated); return theRegion->state(i); }
        virtual const std::vector<_State>& regionState() const { DBG_Assert(region_allocated); return theRegion->state(); }

        virtual MemoryAddress regionTag() const { DBG_Assert(region_allocated); return theRegion->tag(); }

        virtual void setRegionOwner(int32_t new_owner) {
			DBG_Assert(region_allocated);
			rt_cache->updateOwner(*this, new_owner, tagset);
		}
        virtual void setRegionState(const std::vector<_State> &new_state) {
			DBG_Assert(false, ( << "setRegionState() not implemented yet." ));
		}
        virtual void setRegionSharerState(int32_t sharer, boost::dynamic_bitset<uint64_t> &presence, boost::dynamic_bitset<uint64_t> &exclusive) {
			DBG_Assert(false, ( << "setRegionSharerState() not implemented yet." ));
		}
	
	};

	friend class RTLookupResult;

	typedef typename boost::intrusive_ptr<AbstractLookupResult<_State> > AbstractLookup_p;
	typedef AbstractRegionLookupResult<_State> ARLookup;
	typedef typename boost::intrusive_ptr<ARLookup> ARLookup_p;
	typedef typename boost::intrusive_ptr<RTLookupResult> LookupResult_p;

	enum RTReplPolicy_t { SET_LRU, REGION_LRU };

	typedef boost::function<void (MemoryAddress, int, bool)> observer_func_t;
	typedef std::list<observer_func_t>	observer_list_t;

	class RTEvictBuffer : public DirEvictBuffer<_EState> {
	public:

		virtual const AbstractDirEBEntry<_EState>* find(MemoryAddress anAddress) {
			return theDirectory.theBlockEvictBuffer.find(anAddress);
		}

		virtual void insert(MemoryAddress addr, _EState state) { DBG_Assert(false, ( << "insert() NOT supported by RTEvictBuffer." )); }
		virtual void remove(MemoryAddress anAddress) { theDirectory.theBlockEvictBuffer.remove(anAddress); }

		virtual void reserve(int32_t n = 1) {
			theDirectory.theBlockEvictBuffer.reserve(n);
			theDirectory.theERBReserve += n;
		}

		virtual void unreserve(int32_t n = 1) {
			theDirectory.theBlockEvictBuffer.unreserve(n);
			theDirectory.theERBReserve -= n;
		}

		virtual bool empty() const { return ((theDirectory.theERB.size() == 0) && theDirectory.theBlockEvictBuffer.empty()); }
		
		virtual bool full() const {
			return ((((int)theDirectory.theERB.size() + theDirectory.theERBReserve) >= theDirectory.theERBSize) || theDirectory.theBlockEvictBuffer.full());
		}

		virtual bool freeSlotsPending() const { 
//bool ret = theDirectory.theBlockEvictBuffer.freeSlotsPending();
//DBG_(Dev, ( << theDirectory.theName << " - freeSlotsPending() = " << std::boolalpha << ret ));
//			return ret;
			return theDirectory.theBlockEvictBuffer.freeSlotsPending();
		}

		virtual bool idleWorkReady() const {	return theDirectory.theBlockEvictBuffer.idleWorkReady(); }

		virtual const AbstractDirEBEntry<_EState>* oldestRequiringInvalidates() {
//const AbstractDirEBEntry<_EState>* ret = theDirectory.theBlockEvictBuffer.oldestRequiringInvalidates();
//DBG_(Dev, ( << theDirectory.theName << " - oldestRequiringInvalidates() = " << (ret == NULL ? "NULL" : "Non-NULL") ));
//return ret;
			return theDirectory.theBlockEvictBuffer.oldestRequiringInvalidates();
		}

	private:
		RTEvictBuffer(RTDirectory<_State,_EState> &aDir) : DirEvictBuffer<_EState>(0), theDirectory(aDir) {}

		RTDirectory<_State,_EState>& theDirectory;

		friend class RTDirectory<_State,_EState>;
	};

	friend class RTEvictBuffer;

private:
	// Private data
	std::string theName;
	int32_t theNodeId;

	int32_t theBlockSize;
	int32_t theRegionSize;
	int32_t theNumDataSets;
	int32_t theAssociativity;
	int32_t theNumRTSets;
	int32_t theRTAssociativity;

	int32_t theERBSize;
	int32_t theERBReserve;

	int32_t theOrigERBSets;
	int32_t theOrigERBAssoc;

	int32_t theBanks;
	int32_t theTotalBanks;
	int32_t theBankInterleaving;
	int32_t theLocalBankIndex;
	int32_t theGlobalBankIndex;
	int32_t theGroups;
	int32_t theGroupInterleaving;
	int32_t theGroupIndex;
	int32_t theNumSharers;
	int32_t theSkewSet;

	RTReplPolicy_t theReplPolicy;

	block_set_t *theBlocks;
	rt_set_t	*theRVA;
	rt_set_t	theERB;

	std::vector<boost::dynamic_bitset<uint16_t> >	theRegionUsedBits;
	std::vector<boost::dynamic_bitset<uint16_t> >	theBlockUsedBits;

	int32_t theRegionPLRUCounter;
	int32_t theBlockPLRUCounter;

	DirEvictBuffer<_EState>	theBlockEvictBuffer;
	RTEvictBuffer			theRTEvictBuffer;

	int32_t setSkewShift;
	int32_t regionShift, rtSetMidShift, rtSetHighShift;
	uint64_t rtSetLowMask;
	uint64_t rtSetMidMask;
	uint64_t rtSetHighMask;

	uint64_t blockOffsetMask;
	int32_t blockShift, blockSetMidShift, blockSetHighShift;
	uint64_t blockSetLowMask;
	uint64_t blockSetMidMask;
	uint64_t blockSetHighMask;

	uint64_t blockTagMask;
	uint64_t rtTagMask;

	int32_t blockTagShift;
	int32_t theBlocksPerRegion;

	uint64_t rtGlobalSetMask;
	uint64_t globalSetMask;

	int32_t theBankMask, theBankShift;
	int32_t theGroupMask, theGroupShift;

	Flexus::Stat::StatCounter	theRegionReplacements_Empty;
	Flexus::Stat::StatCounter	theRegionReplacements_Unshared;
	Flexus::Stat::StatCounter	theRegionReplacements_Active;

	inline uint64_t get_rt_set(uint64_t addr) {
		//return ( (addr >> regionShift) & rtSetMask );
		if (theSkewSet) {
			addr = MemoryAddress(addr ^ (addr >> setSkewShift));
		}
		return ((addr >> regionShift) & rtSetLowMask) | ((addr >> rtSetMidShift) & rtSetMidMask) | ((addr >> rtSetHighShift) & rtSetHighMask);
	}

	inline uint64_t get_global_rt_set(uint64_t addr) {
		if (theSkewSet) {
			addr = MemoryAddress(addr ^ (addr >> setSkewShift));
		}
		return ((addr >> regionShift) & rtGlobalSetMask);
	}

	inline int64_t get_rt_tag(uint64_t addr) {
		return (addr & rtTagMask);
	}

	inline uint64_t get_block_offset(uint64_t addr) {
		return ( (addr >> blockShift) & blockOffsetMask );
	}

	inline uint64_t get_block_set(uint64_t addr) {
		if (theSkewSet) {
			// need to make sure we don't mess with the block offset bits here
			addr = MemoryAddress(addr ^ ((addr >> setSkewShift) & rtTagMask));
		}
		return ((addr >> blockShift) & blockSetLowMask) | ((addr >> blockSetMidShift) & blockSetMidMask) | ((addr >> blockSetHighShift) & blockSetHighMask);
	}
	inline uint64_t get_global_set(uint64_t addr) {
		if (theSkewSet) {
			addr = MemoryAddress(addr ^ ((addr >> setSkewShift) & rtTagMask));
		}
		return ((addr >> blockShift) & globalSetMask);
	}

	inline uint64_t get_block_tag(uint64_t addr) {
		return (addr & blockTagMask);
	}

	inline uint64_t getRegionNumber(MemoryAddress addr) {
		return (addr >> regionShift);
	}

	inline int32_t getBankIndex(MemoryAddress addr) {
		return (addr >> theBankShift) & theBankMask;
	}

	inline int32_t getGroupIndex(MemoryAddress addr) {
		return (addr >> theGroupShift) & theGroupMask;
	}

	static RTReplPolicy_t string2ReplPolicy(const std::string &policy)
	{
		if (policy == "RegionLRU") {
			return REGION_LRU;
		} else if (policy == "SetLRU") {
			return SET_LRU;
		} else {
			DBG_(Crit, ( << "Unknown RTCache Replacement Policy '" << policy << "'" ));
		}
		return SET_LRU;
	}

	int32_t pickRegionVictimWay(rt_set_t	*rt_set, int32_t set_index) {

		// Loop through regions
		rt_way_iterator iter = rt_set->get<by_way>().find(theRegionPLRUCounter);
		rt_way_iterator end = rt_set->get<by_way>().end();
		theRegionPLRUCounter = (theRegionPLRUCounter + 1) % theRTAssociativity;

		int32_t first_unprotected = -1;
		int32_t first_unused = -1;
		for(int32_t i = 0; i < theRTAssociativity; i++,iter++) {
			if (iter == end) {
				iter = rt_set->get<by_way>().begin();
			}
			if (iter->protect.any()) {
				continue;
			}
			if (iter->valid.none()) {
				return iter->way();
			}
			if (first_unused < 0 && theRegionUsedBits[set_index][iter->way()]) {
				first_unused = iter->way();
			} else if (first_unprotected < 0) {
				first_unprotected = iter->way();
			}
		}

		if (first_unused >= 0) {
			return first_unused;
		} else {
			return first_unprotected;
		}
	}

	void init()
	{
		DBG_(Dev, ( << theName << " - Creating RT Directory: Bsize = " << theBlockSize << ", RSize = " << theRegionSize << ", Sets = " << theNumDataSets << ", Assoc = " << theAssociativity << ", RTSets = " << theNumRTSets << ", RTAssoc = " << theRTAssociativity << ", ERBSize = " << theERBSize ));

		DBG_Assert( theNumDataSets > 0 );
		DBG_Assert( theNumRTSets > 0 );
		DBG_Assert( theAssociativity > 0 );
		DBG_Assert( theRTAssociativity > 0 );
		DBG_Assert( theRegionSize > theBlockSize );
		DBG_Assert( theERBSize > 0 );

		theBlocks = new block_set_t[theNumDataSets];
		theBlockUsedBits.resize(theNumDataSets, boost::dynamic_bitset<uint16_t>(theAssociativity, true));

		theBlocksPerRegion = theRegionSize / theBlockSize;

		DBG_Assert ( (theNumDataSets / theBlocksPerRegion) <= theNumRTSets, ( << "DataSets = " << theNumDataSets << ", RTSets = " << theNumRTSets << ", BlocksPerRegion = " << theBlocksPerRegion ));

		// Make sure everything is a power of 2
		DBG_Assert( ((theNumDataSets - 1) & theNumDataSets) == 0 );
		DBG_Assert( ((theNumRTSets - 1) & theNumRTSets) == 0 );
		DBG_Assert( ((theRegionSize - 1) & theRegionSize) == 0 );
		DBG_Assert( ((theBlockSize - 1) & theBlockSize) == 0 );

        DBG_Assert( ((theBanks - 1) & theBanks) == 0);
        DBG_Assert( ((theBankInterleaving - 1) & theBankInterleaving) == 0);
        DBG_Assert( theBankInterleaving >= theRegionSize );
        DBG_Assert( ((theGroups - 1) & theGroups) == 0);
        DBG_Assert( ((theGroupInterleaving - 1) & theGroupInterleaving) == 0);
        DBG_Assert( theGroupInterleaving >= (theBankInterleaving * theBanks) );

		DBG_Assert( (theBankInterleaving * theBanks) <= theGroupInterleaving, ( << "Invalid interleaving: BI = " 
				<< theBankInterleaving << ", Banks = " << theBanks << ", GI = " << theGroupInterleaving << ", Groups = " << theGroups ));


		theRVA = new rt_set_t[theNumRTSets];
		theRegionUsedBits.resize(theNumRTSets, boost::dynamic_bitset<uint16_t>(theRTAssociativity, true));

        //  Physical address layout:                                                                                                               
        //                                                                                                                                         
        //  +---------+--------------+------+-------------+----------------------------+                                                               
        //  | Tag     | R Index High | Bank | R Index Low | RegionOffset | BlockOffset |                                                               
        //  +---------+--------------+------+-------------+----------------------------+                                                               
        //                                                |<------ setLowShift ------->|                                                               
        //                                  |<----------->|                                                                                            
        //                                     setMaskLow 
		

		// Calculate shifts and masks

        int32_t blockOffsetBits		= log_base2(theBlockSize);
        int32_t regionOffsetBits	= log_base2(theRegionSize);
        int32_t rtIndexBits			= log_base2(theNumRTSets);
        int32_t bIndexBits			= log_base2(theNumDataSets);
        int32_t bankBits			= log_base2(theBanks);
        int32_t bankInterleavingBits    = log_base2(theBankInterleaving);
        int32_t groupBits            = log_base2(theGroups);
        int32_t groupInterleavingBits    = log_base2(theGroupInterleaving);

        int32_t rtLowBits = bankInterleavingBits - regionOffsetBits;
        int32_t rtMidBits = groupInterleavingBits - bankInterleavingBits - bankBits;
        int32_t rtHighBits = rtIndexBits - rtMidBits - rtLowBits;
        if ((rtMidBits + rtLowBits) > rtIndexBits) {
            rtMidBits = rtIndexBits - rtLowBits;
            rtHighBits = 0;
        }

        int32_t bLowBits = bankInterleavingBits - blockOffsetBits;
        int32_t bMidBits = groupInterleavingBits - bankInterleavingBits - bankBits;
        int32_t bHighBits = bIndexBits - bMidBits - bLowBits;
        if ((bMidBits + bLowBits) > bIndexBits) {
            bMidBits = bIndexBits - bLowBits;
            bHighBits = 0;
        }

        DBG_Assert( bHighBits >= 0, ( << "Interleaving is too high for the number of sets." ));

        regionShift     = regionOffsetBits;;
        rtSetLowMask    = ((1 << rtLowBits) - 1);
        rtSetMidMask   = ((1 << rtMidBits) - 1) << rtLowBits;
        rtSetHighMask   = ((1 << rtHighBits) - 1) << (rtMidBits + rtLowBits);
        rtSetMidShift  = bankBits + regionShift;
        rtSetHighShift  = groupBits + bankBits + regionShift;
        rtTagMask       = ~(theRegionSize - 1);

        blockShift      = blockOffsetBits;
        blockTagMask    = ~(theBlockSize - 1);
        blockSetLowMask = (1 << bLowBits) -1;;
        blockSetMidMask    = ((1 << bMidBits) - 1) << bLowBits;
        blockSetHighMask    = ((1 << bHighBits) - 1) << (bMidBits + bLowBits);
        blockSetMidShift   = bankBits + blockOffsetBits;
        blockSetHighShift   = groupBits + bankBits + blockOffsetBits;
        blockOffsetMask = (theBlocksPerRegion - 1);

        globalSetMask = (1 << (bankBits + bIndexBits)) - 1;
        rtGlobalSetMask = (1 << (bankBits + rtIndexBits)) - 1;

        theBankMask = theBanks - 1;
        theBankShift = bankInterleavingBits;
        theGroupMask = theGroups - 1;
        theGroupShift = bankInterleavingBits;

		setSkewShift = 34 - rtIndexBits - bankBits - regionOffsetBits;

		theERBReserve = 0;

DBG_(Verb, ( << "Each RTEntry is " << sizeof(RTEntry) << " bytes, and each BlockEntry is " << sizeof(BlockEntry) << " bytes." ));
		DBG_(Verb, ( << "Created RTDirectory. rShift = " << regionShift << ", bShift = " << blockShift << ", rtIBits = " << rtIndexBits << ", bankBits = " << bankBits << ", skewShift = " << setSkewShift ));

DBG_(Verb, ( << std::hex << "blockOffsetBits = " << blockOffsetBits << ", regionOffsetBits = " << regionOffsetBits << ", bankBits = " << bankBits << ", rtIndexBits = " << rtIndexBits << ", bIndexBits = " << bIndexBits ));
DBG_(Verb, ( << std::hex  << "rtLowBits = " << rtLowBits << ", rtHighBits = " << rtHighBits << ", bLowBits = " << bLowBits << ", bHighBits = " << bHighBits << ", regionShift		 = " << regionShift ));
DBG_(Verb, ( << std::hex << "rtSetLowMask	 = " << rtSetLowMask	 << ", rtSetHighMask	 = " << rtSetHighMask	 << ", rtSetHighShift	 = " << rtSetHighShift	 << ", rtTagMask		 = " << rtTagMask ));
DBG_(Verb, ( << std::hex << "blockShift		 = " << blockShift		 << ", blockTagMask	 = " << blockTagMask	 << ", blockSetLowMask	 = " << blockSetLowMask	 << ", blockSetHighMask	 = " << blockSetHighMask	 << ", blockSetHighShift	 = " << blockSetHighShift	 << ", blockOffsetMask	 = " << blockOffsetMask	 << ", globalSetMask = " << globalSetMask << ", rtGlobalSetMask = " << rtGlobalSetMask ));
	}

	way_iterator replace_block(block_set_t *set, int32_t way, uint64_t tag, const _State& state) {
		way_iterator block = set->get<by_way>().find(way);

		if (block->valid) {
			uint64_t rt_tag = get_rt_tag(block->tag);
			uint64_t rt_set_index = get_rt_set(block->tag);
			int32_t offset = get_block_offset(block->tag);

			rt_set_t	*rt_set	= &(theRVA[rt_set_index]);
			rt_index	*rt		= &(rt_set->get<by_tag>());
			rt_iterator entry	= rt->find(rt_tag);
			rt_iterator end		= rt->end();

			if (entry == end) {
				rt		= &(theERB.get<by_tag>());
				entry	= rt->find(rt_tag);
				end		= rt->end();
				if (entry == end) {
					// Should NEVER get here
					DBG_Assert( false );
				}
			}
			rt->modify(entry, InvalidateRTState(offset));

			// Do the eviction
			DBG_(Verb, ( << "evicting tail block"));
			//DBG_(Trace, Addr(block->tag) ( << theName << " evicting block " << std::hex << block->tag << " in state " << block->state ));

			theBlockEvictBuffer.insert(MemoryAddress(block->tag), block->state);
		}

		set->get<by_way>().modify(block, ReplaceBlock(tag, state));
		return block;
	}

public:

	virtual DirEvictBuffer<_EState>* getEvictBuffer() {
		return &theRTEvictBuffer;
	}
	virtual AbstractRegionEvictBuffer<_EState>* getRegionEvictBuffer() {
		DBG_Assert(false, ( << "getRegionEvictBuffer() not yet implemented."));
		return NULL;
	}

	virtual bool sameSet(MemoryAddress a, MemoryAddress b) {
		// Need to return true if same set in RVA or in BST
		return (get_rt_set(a) == get_rt_set(b)) || (get_block_set(a) == get_block_set(b));
	}

	virtual MemoryAddress getRegion(MemoryAddress addr) {
		return MemoryAddress(get_rt_tag(addr));
	}

	inline int32_t blocksPerRegion() {
		return theBlocksPerRegion;
	}

	// Do we have any idle work?
	virtual bool idleWorkReady() const {
		// In our case, idle work consists of moving blocks from the ERB to the Block Evict Buffer
		// Each move requires a single block lookup

		// To perform idle work, we need an ERB entry, and a free evict buffer entry
		return (theERB.size() > 0 && !theBlockEvictBuffer.full());
	}

	virtual int32_t doIdleWork() {

		// Move one block from ERB to BEB

		const RTEntry &region = theERB.get<by_order>().front();
		uint64_t addr = region.tag();

		bool found_valid_blocks = false;
		bool found_victim = false;
		int32_t i;
		for(i = 0; (i < theBlocksPerRegion) && (!found_victim || !found_valid_blocks); i++,addr+=theBlockSize) {
			block_set_t *tag_set = theBlocks + get_block_set(addr);

			if (region.valid[i]) {
				if (region.protect[i]) {
					found_valid_blocks = true;
					if (found_victim) {
						break;
					}
					continue;
				}
				if (found_victim) {
					break;
				}

				way_iterator block = tag_set->get<by_way>().find(region.ways[i]);
				DBG_Assert( block != tag_set->get<by_way>().end() );

				// Paranoid consistency check
				if (get_rt_tag(block->tag) != region.tag()) {
					DBG_( Crit, ( << theName << " RT tag: " << std::hex << region.tag() << " Block tag: " << block->tag << " offset: " << i << std::dec ));
					tag_iterator t_block = tag_set->get<by_tag>().find(addr);
					tag_iterator t_end = tag_set->get<by_tag>().end();
					if (t_block != t_end) {
						DBG_Assert( !region.valid[i], ( << "found block in way " << t_block->way << " instead of way " << region.ways[i] ));
					} else {
						DBG_Assert( !region.valid[i], ( << "Did not find matching block tag." ));
					}
				} else {
					DBG_(Verb, ( << "Matching region and block tags in region eviction." ));
				}
if (region.block_state[i].noSharers()) {
	DBG_(Crit, ( << "Putting block in EvictBuffer that doesn't require invalidates!!!!"));
}

				theBlockEvictBuffer.insert(MemoryAddress(addr), region.block_state[i]);

				tag_set->get<by_way>().modify(block, ChangeBlockValid(false));
				theERB.get<by_order>().modify(theERB.get<by_order>().begin(), InvalidateRTState(i));

				found_victim = true;
			}
		}
// What if the only block is protected?
		if (i == theBlocksPerRegion && !found_valid_blocks) {
			theERB.get<by_order>().pop_front();
		}
		
		// Only requires one lookup
		// protect+valid bits stored in RVA/ERB entry!!
		return 1;
	}

	RTDirectory(const CMPCacheInfo &theInfo, const std::list< std::pair< std::string, std::string> > &args)
			: theName(theInfo.theName)
			, theNodeId(theInfo.theNodeId)
			, theBlockSize(theInfo.theBlockSize)
			, theRegionSize(-1)
			, theNumDataSets(-1)
			, theAssociativity(-1)
			, theNumRTSets(-1)
			, theRTAssociativity(-1)
			, theERBSize(-1)
			, theERBReserve(-1)
			, theBanks(theInfo.theNumBanks)
			, theBankInterleaving(theInfo.theBankInterleaving)
			, theGroups(theInfo.theNumGroups)
			, theGroupInterleaving(theInfo.theGroupInterleaving)
			, theNumSharers(theInfo.theCores)
			, theSkewSet(false)
			, theBlockEvictBuffer(theInfo.theDirEBSize)
			, theRTEvictBuffer(*this)
			, theRegionReplacements_Empty(theInfo.theName + "-RegionReplacements:Empty")
			, theRegionReplacements_Unshared(theName + "-RegionReplacements:Unshared")
			, theRegionReplacements_Active(theName + "-RegionReplacements:Active")
	{
		theTotalBanks = theBanks * theGroups;
		theGlobalBankIndex = theInfo.theNodeId;
		theLocalBankIndex = theInfo.theNodeId % theBanks;
		theGroupIndex = theInfo.theNodeId / theBanks;

		list<pair<string,string> >::const_iterator iter = args.begin();
		for(; iter != args.end(); iter++) {
			if (strcasecmp(iter->first.c_str(), "sets") == 0) {
				theNumDataSets = strtol(iter->second.c_str(), NULL, 0);
			} else if (strcasecmp(iter->first.c_str(), "total_sets") == 0 || strcasecmp(iter->first.c_str(), "global_sets") == 0) {
                int32_t global_sets = strtol(iter->second.c_str(), NULL, 0);
                theNumDataSets = global_sets / theTotalBanks;
                DBG_Assert( (theNumDataSets * theTotalBanks) == global_sets, ( << "global_sets (" << global_sets << ") is not divisible by number of banks (" << theTotalBanks << ")" ));
			} else if ((strcasecmp(iter->first.c_str(), "assoc") == 0) 
					|| (strcasecmp(iter->first.c_str(), "associativity") == 0)) {
				theAssociativity = strtol(iter->second.c_str(), NULL, 0);
			} else if ((strcasecmp(iter->first.c_str(), "rsize") == 0) 
					|| (strcasecmp(iter->first.c_str(), "region_size") == 0)) {
				theRegionSize = strtol(iter->second.c_str(), NULL, 0);
			} else if ((strcasecmp(iter->first.c_str(), "rtsets") == 0) 
					|| (strcasecmp(iter->first.c_str(), "rt_sets") == 0)) {
				theNumRTSets = strtol(iter->second.c_str(), NULL, 0);
			} else if ((strcasecmp(iter->first.c_str(), "total_rtsets") == 0) 
					|| (strcasecmp(iter->first.c_str(), "total_rt_sets") == 0)
					|| (strcasecmp(iter->first.c_str(), "global_rtsets") == 0)
					|| (strcasecmp(iter->first.c_str(), "global_rt_sets") == 0)) {
                int32_t global_sets = strtol(iter->second.c_str(), NULL, 0);
                theNumRTSets = global_sets / theTotalBanks;
                DBG_Assert( (theNumRTSets * theTotalBanks) == global_sets, ( << "global_rt_sets (" << global_sets << ") is not divisible by number of banks (" << theTotalBanks << ")" ));

			} else if ((strcasecmp(iter->first.c_str(), "rtassoc") == 0) 
					|| (strcasecmp(iter->first.c_str(), "rt_assoc") == 0)
					|| (strcasecmp(iter->first.c_str(), "rtassociativity") == 0)
					|| (strcasecmp(iter->first.c_str(), "rt_associativity") == 0)) {
				theRTAssociativity = strtol(iter->second.c_str(), NULL, 0);
			} else if ((strcasecmp(iter->first.c_str(), "repl") == 0) 
					|| (strcasecmp(iter->first.c_str(), "replacement") == 0)) {
				theReplPolicy = string2ReplPolicy(iter->second);
			} else if ((strcasecmp(iter->first.c_str(), "erb") == 0) 
					|| (strcasecmp(iter->first.c_str(), "erb_size") == 0)) {
				theERBSize = strtol(iter->second.c_str(), NULL, 0);
			} else if ((strcasecmp(iter->first.c_str(), "orig_erb_sets") == 0)) {
				theOrigERBSets = strtol(iter->second.c_str(), NULL, 0);
			} else if ((strcasecmp(iter->first.c_str(), "orig_erb_assoc") == 0)) {
				theOrigERBAssoc = strtol(iter->second.c_str(), NULL, 0);
			} else if ((strcasecmp(iter->first.c_str(), "skew") == 0)
					|| (strcasecmp(iter->first.c_str(), "skew_set") == 0)) {
				theSkewSet = boost::lexical_cast<bool>(iter->second); 
			} else {
				DBG_Assert( false, ( << "Unrecognized directory parameter '" << iter->first << "' = '" << iter->second << "'" ));
			}
		}
		init();
	}
	virtual ~RTDirectory() {
		delete [] theBlocks;
		delete [] theRVA;
	}


	void updateOwner( RTLookupResult &result, int32_t owner, uint64_t tagset) {
		int64_t rt_tag = get_rt_tag(tagset);
		if (!result.region_allocated || result.theRegion->tag() != rt_tag) {
			int	rt_set_index = get_rt_set(tagset);

			rt_set_t	*rt_set	= &(theRVA[rt_set_index]);
			rt_index	*rt		= &(rt_set->get<by_tag>());
			rt_iterator entry	= rt->find(rt_tag);
			rt_iterator end		= rt->end();

			if (entry == end) {
				rt_set  = &theERB;
				rt		= &(theERB.get<by_tag>());
				entry	= rt->find(rt_tag);
				end		= rt->end();
			}

			if (entry != end) {
				rt->modify(entry, ChangeRegionOwner(owner));
			}
		} else {
			result.rt_set->get<by_tag>().modify(result.theRegion, ChangeRegionOwner(owner));
		}
	}

	int32_t getOwner( RTLookupResult &result, uint64_t tagset) {
		uint64_t rt_tag = get_rt_tag(tagset);
		if (!result.region_allocated || result.theRegion->tag() != rt_tag) {
			int	rt_set_index = get_rt_set(tagset);

			rt_set_t	*rt_set	= &(theRVA[rt_set_index]);
			rt_index	*rt		= &(rt_set->get<by_tag>());
			rt_iterator entry	= rt->find(rt_tag);
			rt_iterator end		= rt->end();

			if (entry == end) {
				rt_set  = &theERB;
				rt		= &(theERB.get<by_tag>());
				entry	= rt->find(rt_tag);
				end		= rt->end();
			}

			if (entry != end) {
				return entry->owner;
			} else {
				return -1;
			}
		} else {
			return result.theRegion->owner();
		}
	}

	inline bool isConsistent( uint64_t tagset) {
		return true;
// Fail-safe code to detect bugs in RegionTracker
// It looks like the codes working, so it's turned off to speed things up
#if 0
		block_set_t *block_set = &(theBlocks[get_block_set(tagset, 0)]);
		tag_iterator block = block_set->get<by_tag>().find(get_block_tag(tagset));
		tag_iterator bend   = block_set->get<by_tag>().end();

		uint64_t rt_tag = get_rt_tag(tagset);
		int	rt_set_index = get_rt_set(tagset);
		int32_t offset = get_block_offset(tagset);

		rt_set_t	*rt_set	= &(theRVA[rt_set_index]);
		rt_index	*rt		= &(rt_set->get<by_tag>());
		rt_iterator entry	= rt->find(rt_tag);
		rt_iterator end		= rt->end();

		if (entry == end) {
			rt_set  = &theERB;
			rt		= &(theERB.get<by_tag>());
			entry	= rt->find(rt_tag);
			end		= rt->end();
		}

		if (entry == end) {
			if (block != bend) {
				if (block->state.isValid()) {
					DBG_(Crit, ( << "block " << std::hex << tagset << " region not present but block is valid"));
					return false;
				}
			}
			return true;
		}
		if ((block == bend) && (entry->state[offset].isValid())) {
			DBG_(Crit, Set( (CompName) << theName) Addr(tagset) ( << "block " << std::hex << tagset << " not present but RVA says valid"));
			return false;
		}
		if ((block != bend) && entry->state[offset] != block->state)  {
			DBG_(Crit, Set( (CompName) << theName) Addr(tagset) ( << "block " << std::hex << tagset << " RVA and BST states do not match (" << entry->state[offset] << " != " << block->state << ")" ));
			return false;
		}
		if ((block != bend) && (entry->ways[offset] != block->way) && (block->state.isValid())) {
			DBG_(Crit, Set( (CompName) << theName) Addr(tagset) ( << "block " << std::hex << tagset << " RVA and BST ways do not match (" << entry->ways[offset] << " != " << block->way << ")" ));
			return false;
		}
		return true;
#endif
	}

	virtual AbstractLookup_p lookup(MemoryAddress anAddress) {
		return nativeLookup(anAddress);
	}

	virtual ARLookup_p nativeLookup(MemoryAddress anAddress) {

		uint64_t tagset	= anAddress;
		uint64_t rt_tag	= get_rt_tag(tagset);

		int	rt_set_index	= get_rt_set(tagset);
		int32_t offset			= get_block_offset(tagset);
		bool region_allocated = false;

		rt_set_t	*rt_set	= &(theRVA[rt_set_index]);
		rt_index	*rt		= &(rt_set->get<by_tag>());
		rt_iterator entry	= rt->find(rt_tag);
		rt_iterator end		= rt->end();

		if (entry == end) {
			// No region entry in RVA, check ERB
			rt_set  = &theERB;
			rt		= &(theERB.get<by_tag>());
			entry	= rt->find(rt_tag);
			end		= rt->end();
		}

		if (entry != end) {
			// We have a region entry
			region_allocated = true;

			if (entry->valid[offset]) {
				// The block is valid

				block_set_t *block_set = &(theBlocks[get_block_set(tagset)]);

				tag_iterator t_block = block_set->get<by_tag>().find(get_block_tag(tagset));
				tag_iterator t_end   = block_set->get<by_tag>().end();

				DBG_Assert( t_block != t_end, ( << "tagset 0x" << std::hex << (uint64_t)tagset << ", BTag 0x" << (uint64_t)get_block_tag(tagset) << ", rt_tag 0x" << rt_tag << ", rt_set 0x" << rt_set_index << ", offset " << std::dec << offset ));

				way_iterator block = block_set->get<by_way>().find(t_block->way);
				//way_iterator block = block_set->project<2,tag_iterator>(t_block);

				LookupResult_p ret = new RTLookupResult(true, true, this, tagset, theNumSharers);

				ret->rt_set		= rt_set;
				ret->offset		= offset;
				ret->theRegion		= entry;
				ret->block		= block;
				ret->block_set	= block_set;
				ret->orig_state = block->state;

				DBG_Assert( isConsistent(tagset) );

				updateBlockUsed(get_block_set(ret->tagset), ret->block->way);

				return ret;
			}
		}

		return LookupResult_p( new RTLookupResult(region_allocated, false, this, tagset, theNumSharers)) ;
	}

	void  recordAccess(AbstractLookup_p abs_lookup) {
		RTLookupResult *lookup = dynamic_cast<RTLookupResult*>(abs_lookup.get());
		DBG_Assert(lookup != NULL);
		lookup->updateLRU();
	}

	virtual bool allocate(AbstractLookup_p abs_lookup, MemoryAddress anAddress, const _State &state) {

		RTLookupResult *lookup = dynamic_cast<RTLookupResult*>(abs_lookup.get());
		DBG_Assert(lookup != NULL);

		uint64_t tagset	= anAddress;
		uint64_t rt_tag	= get_rt_tag(tagset);
		int	rt_set_index		= get_rt_set(tagset);
		int32_t offset				= get_block_offset(tagset);

		rt_set_t	*rt_set	= &(theRVA[rt_set_index]);
		rt_index	*rt		= &(rt_set->get<by_tag>());
		rt_iterator entry	= rt->find(rt_tag);
		rt_iterator end		= rt->end();

		if (entry == end) {
			// No region entry in RVA, check ERB
			rt_set  = &theERB;
			rt		= &(theERB.get<by_tag>());
			entry	= rt->find(rt_tag);
			end		= rt->end();
		}

		// We didn't find a region entry in the RVA or the ERB
		if (entry == end) {
			DBG_(Verb, ( << theName << " - Allocating new RVA entry with tag: " << std::hex << rt_tag << std::dec ));
			rt_set	= &(theRVA[rt_set_index]);
			rt		= &(rt_set->get<by_tag>());

			// Need to allocate a new RVA entry.
			// Only do this if we can also allocated a BST entry, e.g., there is at least one un-protected BST entry in the right set.
			block_set_t *block_set = &(theBlocks[get_block_set(tagset)]);
			way_iterator w_iter = block_set->get<by_way>().begin();
			way_iterator w_end = block_set->get<by_way>().end();
			int32_t count = 0;
			for(; w_iter != w_end; count++,w_iter++) {
				if (!w_iter->protect) {
					break;
				}
			}
			if (count == theAssociativity) {
				// No unprotect BST entries, return now before allocating anything
DBG_(Trace, ( << theName << " - no un-protected block ways found while trying to allocate " << std::hex << tagset ));
				return false;
			}

			// Start by adding another way to the set
			int32_t way = rt->size();

			// If the RVA set has too may ways, move one to the ERB
			if (way >= theRTAssociativity) {

				// pick a victim way
				way = pickRegionVictimWay(rt_set,rt_set_index);

				// If there were no un-protected ways, return false
				if (way < 0) {
DBG_(Trace, ( << theName << " - no un-protected ways found while trying to allocate " << std::hex << tagset ));
					return false;
				}

				rt_way_iterator victim_region = rt_set->get<by_way>().find(way);
				theERB.get<by_order>().push_back(*victim_region);
				rt_set->get<by_way>().erase(victim_region);

				// We need to reserve space in the ERB anytime we might need to evict a region
				DBG_Assert( (int)theERB.size() <= theERBSize );

			}

			// Finally, add the new region entry to the RVA set
			std::pair<rt_iterator, bool> temp;
			temp = rt_set->get<by_tag>().insert(RTEntry(theBlocksPerRegion, theNumSharers, rt_tag, way));
			DBG_Assert((int)rt_set->get<by_tag>().size() <= theRTAssociativity);
			DBG_Assert(temp.second);
			entry = temp.first;
			end   = rt->end();

			//allocRegion(anAddress);
		}

		DBG_Assert( entry != end );

		// At this point we have a region entry
		// Start worrying about a block entry

		int32_t block_set_index = get_block_set(tagset);
		block_set_t *block_set = &(theBlocks[block_set_index]);

		if (entry->valid[offset]) {

			DBG_Assert(false, ( << "Tried to allocate a block that's already in the directory." ));

			// The block is valid

			tag_iterator t_block = block_set->get<by_tag>().find(get_block_tag(tagset));
			tag_iterator t_end   = block_set->get<by_tag>().end();

			DBG_Assert( t_block != t_end );

			way_iterator block = block_set->get<by_way>().find(t_block->way);
			//way_iterator block = block_set->project<by_way>(t_block);

			lookup->region_allocated	= true;
			lookup->block_allocated	= true;
			lookup->rt_cache	= this;
			lookup->block_set	= block_set;
			lookup->tagset	= tagset;
			lookup->rt_set	= rt_set;
			lookup->offset	= offset;
			lookup->theRegion	= entry;
			lookup->block	= block;
			lookup->orig_state = block->state;

			DBG_Assert( isConsistent(tagset) );

			DBG_(Trace, Set( (CompName) << theName) Addr(tagset) ( << " Inserting block " << std::hex << tagset << " in set: " << std::hex << get_block_set(tagset) << ", way: " << block->way << "(" << entry->ways[offset] << "), size: " << block_set->size()));

			return true;

		} else {
			// The block is not currently valid, so let's allocate space for it

			// First, find a replacement block
			uint64_t block_tag = get_block_tag(tagset);
			way_iterator block = block_set->get<by_way>().end();
			int32_t way = block_set->size();

			// Try to find a matching tag in the BST
			// This is a somewhat un-realistic optimization since the BST
			// doesn't contain tags, but it simplifies things in the code
			// if we re-use entries like this, and it doesn't result in any
			// un-realistic behaviour unless we care about the specific "way" being used
			// Thus, if we implement a Non-LRU block replacement policy, we might not
			// want to do this.

			tag_iterator t_block = block_set->get<by_tag>().find(block_tag);
			tag_iterator t_end   = block_set->get<by_tag>().end();
			if (t_block != t_end) {
				//block = block_set->project<by_way>(t_block);
				block = block_set->get<by_way>().find(t_block->way);
				way = block->way;
				block_set->get<by_way>().modify(block, ChangeBlockState(state));
				block_set->get<by_way>().modify(block, ChangeBlockValid(true));
DBG_Assert(block->way == way);
				//DBG_(Verb, ( << theName << "Found matching BST tag, new state is " << block->state));
			} else {

				// First, check if the set is full yet, if not simply add the block to the set.
				if (way < theAssociativity) {
					std::pair<way_iterator, bool> result = block_set->get<by_way>().insert(BlockEntry(block_tag, state, way, true));
					if (!result.second) {
						DBG_(Dev, ( << theName << " Failed to insert block. Existing block tag: " << std::hex << result.first->tag << std::dec << ", way: " << result.first->way ));
						DBG_(Dev, ( << theName << " Failed to insert block. New block tag: " << std::hex << block_tag << std::dec << ", way: " << way ));
						//DBG_(Dev, ( << theName << " Was in state " << std::hex << result.first->state << std::dec ));
						DBG_(Dev, ( << theName << " size: " << block_set->get<by_way>().size() ));
						DBG_(Dev, ( << theName << " set: " << std::hex << block_tag ));
						DBG_(Dev, ( << theName << " Inserting block " << std::hex << block_tag << " in set: " << get_block_set(tagset) << ", way: " << way << ", size: " << block_set->size() ));
						DBG_Assert(false);
					}
					block = result.first;
DBG_Assert(block->way == way);
					//DBG_(Trace, ( << theName << " Allocated new block: " << std::hex << tagset << ", state is " << block->state));

				} else {

					// Now look for a replacement

					// Use PLRU policy similar to OpenSparc
					// but prioritize invalid blocks
					
					way_iterator iter = block_set->get<by_way>().find(theBlockPLRUCounter);
					way_iterator end = block_set->get<by_way>().end();
					theBlockPLRUCounter = (theBlockPLRUCounter + 1) % theAssociativity;

					int32_t first_unprotected = -1;
					int32_t first_unused = -1;
					int32_t empty_way = -1;
					for(int32_t i = 0; i < theAssociativity; i++,iter++) {
						if (iter == end) {
							iter = block_set->get<by_way>().begin();
						}
						if (iter->protect) {
							continue;
						}
						if (!iter->valid) {
							empty_way = iter->way;
							break;
						}
						if (first_unused < 0 && !theBlockUsedBits[block_set_index][iter->way]) {
							first_unused = iter->way;
						} else if (first_unprotected < 0) {
							first_unprotected = iter->way;
						}
					}

					if (empty_way >= 0) {
						way = empty_way;
					} else if (first_unused >= 0) {
						way = first_unused;
					} else if (first_unprotected >= 0) {
						way = first_unused;
					} else {
DBG_(Trace, ( << theName << " - no un-protected block ways found while trying to allocate " << std::hex << tagset ));
						return false;
					}
					block = replace_block(block_set, way, block_tag, state);

				}
			}

			DBG_Assert(block != block_set->get<by_way>().end());
			DBG_Assert(way < theAssociativity);

			rt->modify(entry, AllocateBlock(offset, state, way));


			//DBG_(Verb, ( << theName << " RVA state set to: " << entry->block_state[offset] ));

			lookup->region_allocated = true;
			lookup->block_allocated = true;
			lookup->rt_cache 	= this;
			lookup->tagset		= tagset;
			lookup->rt_set		= rt_set;
			lookup->offset		= offset;
			lookup->theRegion		= entry;
			lookup->block		= block;
			lookup->block_set	= block_set;
			lookup->orig_state	= state;

			DBG_(Trace, Set( (CompName) << theName) Addr(tagset) ( << " Inserting block " << std::hex << tagset << " in set: " << std::hex << get_block_set(tagset) << ", way: " << block->way << "(" << entry->ways[offset] << "), size: " << block_set->size()));

			DBG_Assert( isConsistent(tagset) );
			return true;
		}
	}

	void setProtected(RTLookupResult &result, bool val) {
		result.rt_set->get<by_tag>().modify(result.theRegion, ChangeRTProtected(result.offset, val));
		result.block_set->get<by_way>().modify(result.block, ChangeBlockProtected(val));
	}

	void setValid(RTLookupResult &result, bool val) {
		result.rt_set->get<by_tag>().modify(result.theRegion, ChangeRTValid(result.offset, val));
		result.block_set->get<by_way>().modify(result.block, ChangeBlockValid(val));
	}

	inline void updateBlockUsed(int32_t set_index, int32_t way) {

		theBlockUsedBits[set_index][way] = false;
		if (theBlockUsedBits[set_index].none()) {
			theBlockUsedBits[set_index].set();
			theBlockUsedBits[set_index][way] = false;
		}
	}

	inline void updateRegionUsed(int32_t set_index, int32_t way) {

		theRegionUsedBits[set_index][way] = false;
		if (theRegionUsedBits[set_index].none()) {
			theRegionUsedBits[set_index].set();
			theRegionUsedBits[set_index][way] = false;
		}
	}

	void updateState(RTLookupResult &result, const _State &new_state, bool make_MRU, bool make_LRU) {
		DBG_Assert(result.block_allocated);

		//DBG_(Trace, Set( (CompName) << theName ) Addr(result.tagset) ( << "Setting State of block: " << std::hex << result.tagset << " to " << new_state ));
		
		result.rt_set->get<by_tag>().modify(result.theRegion, ChangeRTState(result.offset, new_state));
		result.block_set->get<by_way>().modify(result.block, ChangeBlockState(new_state));

		// Fixup LRU order
		if (make_MRU) {
			updateBlockUsed(get_block_set(result.tagset), result.block->way);
			updateRegionUsed(get_rt_set(result.tagset), result.theRegion->way());
		}
		DBG_Assert(isConsistent(result.tagset));
	}

	void addSharer(RTLookupResult &result, int32_t sharer) {
		DBG_Assert(result.block_allocated);

		//DBG_(Trace, Set( (CompName) << theName ) Addr(result.tagset) ( << "Adding sharer " << sharer << " to block " << std::hex << result.tagset << ". Current state = " << result.block->state ));
		
		result.rt_set->get<by_tag>().modify(result.theRegion, AddRTSharer(result.offset, sharer));
		result.block_set->get<by_way>().modify(result.block, AddBlockSharer(sharer));

		updateBlockUsed(get_block_set(result.tagset), result.block->way);
		updateRegionUsed(get_rt_set(result.tagset), result.theRegion->way());

		DBG_Assert(isConsistent(result.tagset));
	}

	void removeSharer(RTLookupResult &result, int32_t sharer) {
		DBG_Assert(result.block_allocated);

		//DBG_(Trace, Set( (CompName) << theName ) Addr(result.tagset) ( << "Removing sharer " << sharer << " from block " << std::hex << result.tagset << ". Current state = " << result.block->state ));
		
		result.rt_set->get<by_tag>().modify(result.theRegion, RemoveRTSharer(result.offset, sharer));
		result.block_set->get<by_way>().modify(result.block, RemoveBlockSharer(sharer));

		
		if (result.block->state.noSharers()) {
			// If there are no more sharers, remove block from directory
			result.rt_set->get<by_tag>().modify(result.theRegion, InvalidateRTState(result.offset));
			result.block_set->get<by_way>().modify(result.block, ChangeBlockValid(false));
		}

		DBG_Assert(isConsistent(result.tagset));
	}

	void setSharer(RTLookupResult &result, int32_t sharer) {
		DBG_Assert(result.block_allocated);

		//DBG_(Trace, Set( (CompName) << theName ) Addr(result.tagset) ( << "Setting sharer " << sharer << " for block " << std::hex << result.tagset << ". Current state = " << result.block->state ));
		
		result.rt_set->get<by_tag>().modify(result.theRegion, SetRTSharer(result.offset, sharer));
		result.block_set->get<by_way>().modify(result.block, SetBlockSharer(sharer));

		updateBlockUsed(get_block_set(result.tagset), result.block->way);
		updateRegionUsed(get_rt_set(result.tagset), result.theRegion->way());

		DBG_Assert(isConsistent(result.tagset));
	}

	int32_t regionProbeOwner(MemoryAddress anAddress) {

		uint64_t tagset = anAddress;
		uint64_t rt_tag = get_rt_tag(tagset);
		int	rt_set_index = get_rt_set(tagset);

		rt_set_t	*rt_set	= &(theRVA[rt_set_index]);
		rt_index	*rt		= &(rt_set->get<by_tag>());
		rt_iterator entry	= rt->find(rt_tag);
		rt_iterator end		= rt->end();

		if (entry == end) {
			rt_set  = &theERB;
			rt		= &(theERB.get<by_tag>());
			entry	= rt->find(rt_tag);
			end		= rt->end();
		}

		int32_t owner = -1;
		if (entry != end) {
			owner =  entry->owner;
		}

		DBG_(Verb, ( << theName << ": regionProbeOwner(" << std::hex << tagset << std::dec << ") = " << owner ));
		return owner;
	}
	bool regionSetOwner(MemoryAddress anAddress, int32_t owner) {

		uint64_t tagset = anAddress;
		uint64_t rt_tag = get_rt_tag(tagset);
		int	rt_set_index = get_rt_set(tagset);

		rt_set_t	*rt_set	= &(theRVA[rt_set_index]);
		rt_index	*rt		= &(rt_set->get<by_tag>());
		rt_iterator entry	= rt->find(rt_tag);
		rt_iterator end		= rt->end();

		if (entry == end) {
			rt_set  = &theERB;
			rt		= &(theERB.get<by_tag>());
			entry	= rt->find(rt_tag);
			end		= rt->end();
		}

		if (entry != end) {
			rt->modify(entry, ChangeRegionOwner(owner));
			return true;
		}

		DBG_(Verb, ( << theName << ": regionSetOwner(" << std::hex << tagset << std::dec << ") = " << owner << " FAILED!"));
		return false;
	}

	boost::dynamic_bitset<uint64_t> regionProbePresence(MemoryAddress anAddress) {

		boost::dynamic_bitset<uint64_t> ret(theBlocksPerRegion, false);

		uint64_t tagset = anAddress;
		uint64_t rt_tag = get_rt_tag(tagset);
		int	rt_set_index = get_rt_set(tagset);

		rt_set_t	*rt_set	= &(theRVA[rt_set_index]);
		rt_index	*rt		= &(rt_set->get<by_tag>());
		rt_iterator entry	= rt->find(rt_tag);
		rt_iterator end		= rt->end();

		if (entry == end) {
			rt_set  = &theERB;
			rt		= &(theERB.get<by_tag>());
			entry	= rt->find(rt_tag);
			end		= rt->end();
		}

		if (entry != end) {
			entry->getPresence(ret);
		}

		DBG_(Verb, ( << theName << ": regionProbePresence(" << std::hex << tagset << std::dec << ") = " << ret ));
		return ret;
	}
	const std::vector<_State>* regionProbeState(MemoryAddress anAddress) {

		uint64_t tagset = anAddress;
		uint64_t rt_tag = get_rt_tag(tagset);
		int	rt_set_index = get_rt_set(tagset);

		rt_set_t	*rt_set	= &(theRVA[rt_set_index]);
		rt_index	*rt		= &(rt_set->get<by_tag>());
		rt_iterator entry	= rt->find(rt_tag);
		rt_iterator end		= rt->end();

		if (entry == end) {
			rt_set  = &theERB;
			rt		= &(theERB.get<by_tag>());
			entry	= rt->find(rt_tag);
			end		= rt->end();
		}

		//DBG_Assert(entry != end);
		if (entry == end) {
			return NULL;
		}

		DBG_(Verb, ( << theName << ": regionProbeState(" << std::hex << tagset << std::dec << ")" ));
		return &(entry->state);
	}

	bool saveState( std::ostream & s ) {
		return false;
	}


  bool loadState( std::istream & s) {

	boost::archive::binary_iarchive ia(s);

    int32_t totalRTSets = theNumRTSets * theTotalBanks;
    int32_t global_set = 0;
    int32_t local_set = 0;
    MemoryAddress addr(0);

	uint64_t set_count = 0;
	uint32_t associativity = 0;

	ia >> set_count;
	ia >> associativity;

	DBG_Assert( set_count == (uint64_t)totalRTSets, ( << "Error loading directory state. Flexpoint contains " << set_count << " region sets but simulator configured for " << totalRTSets << " region sets." ));
	DBG_Assert( associativity == (uint64_t)theRTAssociativity, ( << "Error loading directory state. Flexpoint contains " << associativity << "-way region sets but simulator configured for " << theRTAssociativity << "-way region sets." ));

    for(; global_set < totalRTSets; global_set++,addr+=theBlockSize) {
        if ((getBankIndex(addr) == theLocalBankIndex) && (getGroupIndex(addr) == theGroupIndex)) {
			std::vector<SerializableRTDirEntry> temp_rt_set(theRTAssociativity, SerializableRTDirEntry(theBlocksPerRegion));
			// Read in set information
			ia >> temp_rt_set;
			std::vector<uint16_t> blocks(theRegionUsedBits[0].num_blocks());
			ia & blocks;	// PLRU Bits
			from_block_range(blocks.begin(), blocks.end(), theRegionUsedBits[local_set]);
			blocks.clear();

			std::vector<uint16_t> not_valid_bits(theRegionUsedBits[0].num_blocks());
			boost::dynamic_bitset<uint16_t> temp_bitset(theRTAssociativity);

			ia & not_valid_bits;	// !valid bits
			from_block_range(not_valid_bits.begin(), not_valid_bits.end(), temp_bitset);
			not_valid_bits.clear();

			// Copy region entries for valid ways
			int32_t valid_way = 0;
			for(int32_t way = 0; way < theRTAssociativity; way++) {
				if (!temp_bitset[way]) {
					DBG_(Verb, ( << "Adding Region 0x" << std::hex << temp_rt_set[way].tag() << " to set " << local_set ));

					std::pair<rt_iterator,bool> result;
					result = theRVA[local_set].get<by_tag>().insert(RTEntry(theBlocksPerRegion, theNumSharers, temp_rt_set[way].tag(), valid_way, temp_rt_set[way].owner));
					DBG_Assert(result.second);

					if (valid_way != way) {
						theRegionUsedBits[local_set][valid_way] = theRegionUsedBits[local_set][way];
					}

					valid_way++;
				} 
			}
			temp_rt_set.clear();
			local_set++;
		} else {
			std::vector<SerializableRTDirEntry> temp_rt_set(theRTAssociativity, SerializableRTDirEntry(theBlocksPerRegion));
			// Read in set and discard
			ia >> temp_rt_set;
			temp_rt_set.clear();

			std::vector<uint16_t> blocks(theRegionUsedBits[0].num_blocks());
			ia & blocks;	// PLRU Bits
			blocks.clear();
			std::vector<uint16_t> not_valid_bits(theRegionUsedBits[0].num_blocks());
			ia & not_valid_bits;	// !valid bits
			not_valid_bits.clear();
		}
	}


	// Load in the ERB
	// Timing simulator uses fully-associative ERB, trace uses set-associative
	// Ideally, we have one set-per directory bank so this is simple
	// for AtMemory configurations, there are fewer banks, so multiple sets are loaded into the same ERB

	DBG_Assert( (theOrigERBSets * theOrigERBAssoc) == (theERBSize * theTotalBanks) , ( << "OSets = " << theOrigERBSets << ", OAssoc = " << theOrigERBAssoc << ", ERBSize = " << theERBSize ));

	ia >> set_count;
	ia >> associativity;
	DBG_Assert( set_count == (uint64_t)theOrigERBSets, ( << "Error loading directory state. Flexpoint contains " << set_count << " erb sets but simulator configured for " << theOrigERBSets << " erb sets." ));
	DBG_Assert( associativity == (uint64_t)theOrigERBAssoc, ( << "Error loading directory state. Flexpoint contains " << associativity << "-way ERB sets but simulator configured for " << theOrigERBAssoc << "-way ERB sets." ));

	global_set = 0;
	local_set = 0;
	addr = MemoryAddress(0);
	for(; global_set < theOrigERBSets; global_set++,addr+=theBlockSize) {
        if ((getBankIndex(addr) == theLocalBankIndex) && (getGroupIndex(addr) == theGroupIndex)) {
			std::list<SerializableRTDirEntry> erb_set;
			// Read in set information
			ia >> erb_set;

			// Copy each entry to ERB
			std::list<SerializableRTDirEntry>::iterator iter = erb_set.begin();
			std::list<SerializableRTDirEntry>::iterator end = erb_set.end();
			for(int32_t way = 0; iter != end; iter++,way++) {
				theERB.get<by_order>().push_back(RTEntry(theBlocksPerRegion, theNumSharers, iter->tag(), way, iter->owner));
			}
			erb_set.clear();
		} else {
			std::list<SerializableRTDirEntry> erb_set;
			// Read in set and discard
			ia >> erb_set;
			erb_set.clear();
		}
	}

    int32_t totalDataSets = theNumDataSets * theTotalBanks;
    global_set = 0;
    local_set = 0;
    addr = MemoryAddress(0);

	ia >> set_count;
	ia >> associativity;

	DBG_Assert( set_count == (uint64_t)totalDataSets, ( << "Error loading directory state. Flexpoint contains " << set_count << " sets but simulator configured for " << totalDataSets << " sets." ));
	DBG_Assert( associativity == (uint64_t)theAssociativity, ( << "Error loading directory state. Flexpoint contains " << associativity << "-way sets but simulator configured for " << theAssociativity << "-way sets." ));


	StdDirEntrySerializer serializer;
    for(; global_set < totalDataSets; global_set++,addr+=theBlockSize) {
        if ((getBankIndex(addr) == theLocalBankIndex) && (getGroupIndex(addr) == theGroupIndex)) {
			std::vector<StdDirEntrySerializer> block_entries(theAssociativity);
			for(int32_t way = 0; way < theAssociativity; way++) {
				ia >> block_entries[way];
			}
			std::vector<uint16_t> blocks(theBlockUsedBits[0].num_blocks());
			ia & blocks;	// PLRU Bits
			from_block_range(blocks.begin(), blocks.end(), theBlockUsedBits[local_set]);
			blocks.clear();

			boost::dynamic_bitset<uint16_t> temp_bitset(theAssociativity);
			std::vector<uint16_t> not_valid_bits(temp_bitset.num_blocks());

			ia & not_valid_bits;	// !valid bits
			from_block_range(not_valid_bits.begin(), not_valid_bits.end(), temp_bitset);
			not_valid_bits.clear();

			int32_t valid_way = 0;
			for(int32_t way = 0; way < theAssociativity; way++) {
				if (!temp_bitset[way]) {

					std::pair<tag_iterator, bool> result;
					result = theBlocks[local_set].get<by_tag>().insert(BlockEntry(block_entries[way].tag, _State(theNumSharers), valid_way));

					DBG_Assert(result.second);

					_State temp_state(theNumSharers);
					temp_state = block_entries[way].state;
					theBlocks[local_set].get<by_tag>().modify(result.first, ChangeBlockValid(true));
					theBlocks[local_set].get<by_tag>().modify(result.first, ChangeBlockState(temp_state));

					// Need to enter state in RVA or ERB entry for this block
					uint64_t rt_tag	= get_rt_tag(block_entries[way].tag);
					int	rt_set_index			= get_rt_set(block_entries[way].tag);
					int32_t offset					= get_block_offset(block_entries[way].tag);

					rt_set_t	*rt_set	= &(theRVA[rt_set_index]);
					rt_index	*rt		= &(rt_set->get<by_tag>());
					rt_iterator entry	= rt->find(rt_tag);
					rt_iterator end		= rt->end();
					if (entry == end) {
						rt_set	= &theERB;
						rt		= &(rt_set->get<by_tag>());
						entry	= rt->find(rt_tag);
						end		= rt->end();
						DBG_Assert(entry != end, ( << "No region found for block 0x" << std::hex << block_entries[way].tag << " (region 0x" << rt_tag << ") rt_set = " << rt_set_index ));
					}
					rt->modify(entry, AllocateBlock(offset, temp_state, valid_way));

					valid_way++;
				}
			}
			temp_bitset.clear();
			block_entries.clear();
			local_set++;
		} else {
			for(int32_t way = 0; way < theAssociativity; way++) {
				// Read and discard entries for other banks
				ia >> serializer;
			}
			std::vector<uint16_t> blocks(theBlockUsedBits[0].num_blocks());
			ia & blocks;	// PLRU Bits
			blocks.clear();
			std::vector<uint16_t> not_valid_bits(theBlockUsedBits[0].num_blocks());
			ia & not_valid_bits;	// !valid bits
			not_valid_bits.clear();
		}
	}
	
	DBG_(Verb, ( << theName << " - finished loading state, returning true." ));
	// If we get here, we must have succeeded
	return true;
  }

};

}; // namespace nCMPCache
	
#endif // __RT_DIRECTORY_HPP__

