// DO-NOT-REMOVE begin-copyright-block 
//
// Redistributions of any form whatsoever must retain and/or include the
// following acknowledgment, notices and disclaimer:
//
// This product includes software developed by Carnegie Mellon University.
//
// Copyright 2010 by Mohammad Alisafaee, Eric Chung, Michael Ferdman, Brian Gold,
// Jangwoo Kim, Pejman Lotfi-Kamran, Onur Kocberber, Ippokratis Pandis, 
// Djordje Jevdjic, Jared Smolens, Stephen Somogyi, Evangelos Vlachos, 
// Stavros Volos, Jason Zebchuk, Babak Falsafi, Nikos Hardavellas and Tom Wenisch 
// for the SimFlex Project, Computer Architecture Lab at Carnegie Mellon,
// Carnegie Mellon University.
//
// For more information, see the SimFlex project website at:
//   http://www.ece.cmu.edu/~simflex
//
// You may not use the name "Carnegie Mellon University" or derivations
// thereof to endorse or promote products derived from this software.
//
// If you modify the software you must place a notice on or within any
// modified version provided or made available to any third party stating
// that you have modified the software.  The notice shall include at least
// your name, address, phone number, email address and the date and purpose
// of the modification.
//
// THE SOFTWARE IS PROVIDED "AS-IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER
// EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT LIMITED TO ANY WARRANTY
// THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS OR BE ERROR-FREE AND ANY
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
// TITLE, OR NON-INFRINGEMENT.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
// BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO DIRECT, INDIRECT,
// SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM, OR IN
// ANY WAY CONNECTED WITH THIS SOFTWARE (WHETHER OR NOT BASED UPON WARRANTY,
// CONTRACT, TORT OR OTHERWISE).
//
// DO-NOT-REMOVE end-copyright-block   

#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>

#include <core/debug/debug.hpp>

#include <components/FastCMPCache/AbstractDirectory.hpp>
#include <components/FastCMPCache/SharingVector.hpp>
#include <components/FastCMPCache/AbstractProtocol.hpp>
#include <components/FastCMPCache/BlockDirectoryEntry.hpp>
#include <ext/hash_map>

#include <boost/tuple/tuple.hpp>
#include <boost/dynamic_bitset.hpp>
#include <boost/bind.hpp>
#include <boost/lambda/lambda.hpp>

#include <list>
#include <vector>

#include <boost/serialization/serialization.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/list.hpp>

#include <components/Common/Serializers.hpp>
#include <components/Common/Util.hpp>


using nCommonSerializers::StdDirEntrySerializer;
using nCommonUtil::log_base2;

namespace nFastCMPCache {

typedef BlockDirectoryEntry RTBlockEntry;

typedef nCommonSerializers::SerializableRTDirEntry RTDirectoryEntry;

class RTDirectory : public AbstractDirectory {
private:
	int32_t theRegionSets;
	int32_t theRegionAssoc;

	int32_t theERBSets;
	int32_t theERBAssoc;

	int32_t theBlockSets;
	int32_t theBlockAssoc;

	int32_t theRegionSize;
	int32_t theBlocksPerRegion;
	
	uint64_t	theBlockTagMask;
	uint64_t	theRegionTagMask;

	int32_t theOffsetShift;
	int32_t theOffsetMask;

	int32_t theBlockSetShift;
	int32_t theBlockSetMask;
	int32_t theRegionSetShift;
	int32_t theRegionSetMask;
	int32_t theERBSetMask;
	bool theSkewSet;
	int32_t theSkewShift;
	bool theSkewOffset;

	enum {
		ePLRUPolicy,
		eMinPolicy
	} theRegionReplPolicy, theBlockReplPolicy;

	std::vector<std::vector<RTDirectoryEntry> >		theRegionArray;
	std::vector<std::vector<RTBlockEntry> >	theBlockArray;

	std::vector<boost::dynamic_bitset<uint16_t> >	theRegionPLRUBits;
	std::vector<boost::dynamic_bitset<uint16_t> >	theRegionNotValidBits;

	std::vector<boost::dynamic_bitset<uint16_t> >	theBlockPLRUBits;
	std::vector<boost::dynamic_bitset<uint16_t> >	theBlockNotValidBits;

	std::vector<std::list<RTDirectoryEntry> >	theERBArray;

	RTDirectory() : AbstractDirectory(),theRegionReplPolicy(ePLRUPolicy),theBlockReplPolicy(ePLRUPolicy) {};

	virtual void initialize(const std::string &aName) {

		AbstractDirectory::initialize(aName);
		
		DBG_Assert( theBlockSets > 0 );
		DBG_Assert( theBlockAssoc > 0 );
		DBG_Assert( theRegionSets > 0 );
		DBG_Assert( theRegionAssoc > 0 );
		DBG_Assert( theERBSets > 0 );
		DBG_Assert( theRegionSize > 0 );

		theBlocksPerRegion = theRegionSize / theBlockSize;
		DBG_Assert( (theBlockSize * theBlocksPerRegion) == theRegionSize );

		theRegionArray.resize(theRegionSets,
				std::vector<RTDirectoryEntry>(	theRegionAssoc, RTDirectoryEntry(theBlocksPerRegion) ));

		theERBArray.resize(theERBSets);

		theBlockArray.resize(theBlockSets,
				std::vector<RTBlockEntry>(	theBlockAssoc, RTBlockEntry() ));

		theBlockPLRUBits.resize(theBlockSets, boost::dynamic_bitset<uint16_t>(theBlockAssoc));
		theBlockNotValidBits.resize(theBlockSets, boost::dynamic_bitset<uint16_t>(theBlockAssoc));

		std::for_each(theBlockPLRUBits.begin(), theBlockPLRUBits.end(), boost::bind(&boost::dynamic_bitset<uint16_t>::set, _1));
		std::for_each(theBlockNotValidBits.begin(), theBlockNotValidBits.end(), boost::bind(&boost::dynamic_bitset<uint16_t>::set, _1));

		theRegionPLRUBits.resize(theRegionSets, boost::dynamic_bitset<uint16_t>(theRegionAssoc));
		theRegionNotValidBits.resize(theRegionSets, boost::dynamic_bitset<uint16_t>(theRegionAssoc));

		std::for_each(theRegionPLRUBits.begin(), theRegionPLRUBits.end(), boost::bind(&boost::dynamic_bitset<uint16_t>::set, _1));
		std::for_each(theRegionNotValidBits.begin(), theRegionNotValidBits.end(), boost::bind(&boost::dynamic_bitset<uint16_t>::set, _1));

		theBlockSetMask = theBlockSets - 1;
		theRegionSetMask = theRegionSets - 1;
		theERBSetMask = theERBSets - 1;

		theBlockSetShift = log_base2(theBlockSize);
		theRegionSetShift = log_base2(theRegionSize);

		theOffsetShift = log_base2(theBlockSize);
		theOffsetMask = theBlocksPerRegion - 1;

		theRegionTagMask = ~(((uint64_t)theRegionSize)-1);
		theBlockTagMask = ~(((uint64_t)theBlockSize)-1);

		theSkewShift = (34 - log_base2(theRegionSets)) - theRegionSetShift;
	}

	inline int32_t get_block_set(PhysicalMemoryAddress addr) {
		if (theSkewSet) {
			if (theSkewOffset) {
				return ( ((uint64_t)addr >> theBlockSetShift) ^ ( (uint64_t)addr >> (theSkewShift + theBlockSetShift)) ) & theBlockSetMask;
			} else {
				return ( ((uint64_t)addr >> theBlockSetShift) ^ (( (uint64_t)addr >> (theSkewShift + theBlockSetShift)) ) & ~theOffsetMask) & theBlockSetMask;
			}
		} else {
			return ((uint64_t)addr >> theBlockSetShift) & theBlockSetMask;
		}
	}

	inline int32_t get_region_set(PhysicalMemoryAddress addr) {
		if (theSkewSet) {
			return ( ((uint64_t)addr >> theRegionSetShift) ^ ( (uint64_t)addr >> (theSkewShift + theRegionSetShift)) ) & theRegionSetMask;
		} else {
			return ((uint64_t)addr >> theRegionSetShift) & theRegionSetMask;
		}
	}

	inline int32_t get_erb_set(PhysicalMemoryAddress addr) {
		if (theSkewSet) {
			return ( ((uint64_t)addr >> theRegionSetShift) ^ ( (uint64_t)addr >> (theSkewShift + theRegionSetShift)) ) & theERBSetMask;
		} else {
			return ((uint64_t)addr >> theRegionSetShift) & theERBSetMask;
		}
	}

	inline int32_t get_offset(PhysicalMemoryAddress addr) {
		if (theSkewOffset) {
			return (((uint64_t)addr ^ ((uint64_t)addr >> theSkewShift)) >> theOffsetShift) & theOffsetMask;
		} else {
			return ((uint64_t)addr >> theOffsetShift) & theOffsetMask;
		}
	}

	inline PhysicalMemoryAddress get_region_tag(PhysicalMemoryAddress addr) {
		return PhysicalMemoryAddress((uint64_t)addr & theRegionTagMask);
	}
	inline PhysicalMemoryAddress get_block_tag(PhysicalMemoryAddress addr) {
		return PhysicalMemoryAddress((uint64_t)addr & theBlockTagMask);
	}

protected:

    virtual void addSharer(int32_t index, AbstractEntry_p dir_entry, PhysicalMemoryAddress address) {
		//RTBlockEntry *my_entry = dynamic_cast<RTBlockEntry*>(dir_entry);
		RTBlockEntry *my_entry(&(dynamic_cast<BlockEntryWrapper*>(dir_entry.get())->block));
		
		DBG_Assert(my_entry != NULL);
		my_entry->addSharer(index);
	}

    virtual void addExclusiveSharer(int32_t index, AbstractEntry_p dir_entry, PhysicalMemoryAddress address) {
		//RTBlockEntry *my_entry = dynamic_cast<RTBlockEntry*>(dir_entry);
		RTBlockEntry *my_entry(&(dynamic_cast<BlockEntryWrapper*>(dir_entry.get())->block));
		DBG_Assert(my_entry != NULL);
		my_entry->addSharer(index);
		my_entry->makeExclusive(index);
	}

    virtual void failedSnoop(int32_t index, AbstractEntry_p dir_entry, PhysicalMemoryAddress address) {
		// We track precise sharers, so a failed snoop means we remove a sharer
		removeSharer(index, dir_entry, address);
	}

    virtual void removeSharer(int32_t index, AbstractEntry_p dir_entry, PhysicalMemoryAddress address) {
		//RTBlockEntry *my_entry = dynamic_cast<RTBlockEntry*>(dir_entry);
		RTBlockEntry *my_entry(&(dynamic_cast<BlockEntryWrapper*>(dir_entry.get())->block));
		if (my_entry == NULL) {
			return;
		}
		my_entry->removeSharer(index);
	}

    virtual void makeSharerExclusive(int32_t index, AbstractEntry_p dir_entry, PhysicalMemoryAddress address) {
		//RTBlockEntry *my_entry = dynamic_cast<RTBlockEntry*>(dir_entry);
		RTBlockEntry *my_entry(&(dynamic_cast<BlockEntryWrapper*>(dir_entry.get())->block));
		if (my_entry == NULL) {
			return;
		}
		// Make it exclusive
		my_entry->makeExclusive(index);
	}

	inline void updatePLRU(boost::dynamic_bitset<uint16_t> &bits, int32_t way) {
		bits[way] = false;
		if (bits.none()) {
			bits.set();
			bits[way] = false;
		}
	}
	inline void updateRegionLRU(int32_t set, int32_t way) {
		updatePLRU(theRegionPLRUBits[set], way);
	}
	inline void updateBlockLRU(int32_t set, int32_t way) {
		updatePLRU(theBlockPLRUBits[set], way);
	}

	RTDirectoryEntry* findEntry(PhysicalMemoryAddress addr) {

		// First, look for a region entry in the region array
		int64_t region_tag(get_region_tag(addr));
		int32_t region_set_index = get_region_set(addr);
		int32_t region_way = 0;
		for(; region_way < theRegionAssoc; region_way++) {
			if (theRegionArray[region_set_index][region_way].tag() == region_tag) {
				return &(theRegionArray[region_set_index][region_way]);
			}
		}

		// If it's not in the region array, look in the erb
		if (region_way == theRegionAssoc) {
			int32_t erb_set_index = get_erb_set(addr);
			std::list<RTDirectoryEntry>::iterator erb_iter;
			for(erb_iter = theERBArray[erb_set_index].begin(); erb_iter != theERBArray[erb_set_index].end(); erb_iter++) {
				if (erb_iter->tag() == region_tag) {
					return &(*erb_iter);
				}
			}
		}
		return NULL;
	}

	void removeRegion(RTDirectoryEntry *rt_entry) {
		PhysicalMemoryAddress addr(rt_entry->tag());

		// First, look for a region entry in the region array
		int64_t region_tag(get_region_tag(addr));
		int32_t region_set_index = get_region_set(addr);
		int32_t region_way = 0;
		for(; region_way < theRegionAssoc; region_way++) {
			if (theRegionArray[region_set_index][region_way].tag() == region_tag) {
				theRegionNotValidBits[region_set_index][region_way] = true;
				return;
			}
		}

		// If it's not in the region array, look in the erb
		if (region_way == theRegionAssoc) {
			int32_t erb_set_index = get_erb_set(addr);
			std::list<RTDirectoryEntry>::iterator erb_iter;
			for(erb_iter = theERBArray[erb_set_index].begin(); erb_iter != theERBArray[erb_set_index].end(); erb_iter++) {
				if (erb_iter->tag() == region_tag) {
					theERBArray[erb_set_index].erase(erb_iter);
					return;
				}
			}
		}
	}

	int32_t selectMinBlock(int32_t set_index) {
		std::vector<BlockDirectoryEntry> &set = theBlockArray[set_index];
		int32_t min_way = -1;
		int32_t min_sharers = INT_MAX;
		
		for(int32_t way = 0; way < theBlockAssoc; way++) {
			int32_t count = set[way].sharers().countSharers();
			if (count < min_sharers) {
				min_sharers = count;
				min_way = way;
			}
		}
		DBG_Assert(min_way >= 0);
		return min_way;
	}

	int32_t selectMinRegion(int32_t set_index) {
		std::vector<RTDirectoryEntry> &set = theRegionArray[set_index];

		int32_t min_way = -1;
		int32_t min_blocks = theBlocksPerRegion+1;
		int32_t min_sharers = INT_MAX;
		for(int32_t way = 0; way < theRegionAssoc; way++) {
			// Count num blocks in region
			int32_t bcount = 0;
			int32_t max_sharers = 0;
			PhysicalMemoryAddress addr(set[way].tag());
			for(int32_t i = 0; i < theBlocksPerRegion; i++,addr+=theBlockSize) {
				if (set[way][i] >= 0) {
					bcount++;
					BlockDirectoryEntry &block = theBlockArray[get_block_set(addr)][set[way][i]];
					int32_t sharers = block.sharers().countSharers();
					if (sharers > max_sharers) {
						max_sharers = sharers;
					}
				}
			}
			if ((bcount < min_blocks) 
					|| (bcount == min_blocks && max_sharers < min_sharers)) {
				min_way = way;
				min_blocks = bcount;
				min_sharers = max_sharers;
			}
		}
		DBG_Assert(min_way >= 0);
		return min_way;
	}

	RTBlockEntry* findBlockEntry(PhysicalMemoryAddress addr) {

		// First, look for a region entry in the region array
		int64_t region_tag(get_region_tag(addr));
		int32_t region_set_index = get_region_set(addr);
		int32_t region_way = 0;
		for(; region_way < theRegionAssoc; region_way++) {
			if (!theRegionNotValidBits[region_set_index][region_way] && theRegionArray[region_set_index][region_way].tag() == region_tag) {
				break;
			}
		}

		RTDirectoryEntry *rt_entry = NULL;

		// If it's not in the region array, look in the erb
		if (region_way == theRegionAssoc) {
			int32_t erb_set_index = get_erb_set(addr);
			std::list<RTDirectoryEntry>::iterator erb_iter;
			for(erb_iter = theERBArray[erb_set_index].begin(); erb_iter != theERBArray[erb_set_index].end(); erb_iter++) {
				if (erb_iter->tag() == region_tag) {
					break;
				}
			}
			if (erb_iter == theERBArray[erb_set_index].end()) {
				// Not in the RVA or ERB, return NULL
				return NULL;
				
			} else {
				rt_entry = &(*erb_iter);
			}
		} else {
			rt_entry = &(theRegionArray[region_set_index][region_way]);
		}

		DBG_Assert(rt_entry != NULL);

		// Okay, we have a region entry, now we need to find a block entry

		int32_t offset = get_offset(addr);
		int32_t way = (*rt_entry)[offset];
		int32_t block_set = get_block_set(addr);
		if (way < 0) {
			return NULL;
		}

		return &(theBlockArray[block_set][way]);
	}
	RTBlockEntry* findOrCreateEntry(PhysicalMemoryAddress addr, bool update_lru, std::list<boost::function<void(void)> > &actions) {

		// First, look for a region entry in the region array
		int64_t region_tag(get_region_tag(addr));
		int32_t region_set_index = get_region_set(addr);
		int32_t region_way = 0;
		for(; region_way < theRegionAssoc; region_way++) {
			if (!theRegionNotValidBits[region_set_index][region_way] && theRegionArray[region_set_index][region_way].tag() == region_tag) {
				break;
			}
		}

		RTDirectoryEntry *rt_entry = NULL;

		// If it's not in the region array, look in the erb
		if (region_way == theRegionAssoc) {
			int32_t erb_set_index = get_erb_set(addr);
			std::list<RTDirectoryEntry>::iterator erb_iter;
			for(erb_iter = theERBArray[erb_set_index].begin(); erb_iter != theERBArray[erb_set_index].end(); erb_iter++) {
				if (erb_iter->tag() == region_tag) {
					break;
				}
			}
			if (erb_iter == theERBArray[erb_set_index].end()) {
				// No ERB entry found, need to allocate the region
				
				// First, look for invalid RegionEntry
				boost::dynamic_bitset<uint16_t>::size_type pos = theRegionNotValidBits[region_set_index].find_first();
				if (pos == boost::dynamic_bitset<uint16_t>::npos) {
					// Use PLRU bits to find victim
					if (theRegionReplPolicy == ePLRUPolicy) {
						pos = theRegionPLRUBits[region_set_index].find_first();
					} else {
						pos = selectMinRegion(region_set_index);
					}
					DBG_Assert(pos != boost::dynamic_bitset<uint16_t>::npos);
					theERBArray[erb_set_index].push_front(theRegionArray[region_set_index][pos]);

DBG_(Trace, ( << "Moving region to ERB " << std::hex << theERBArray[erb_set_index].front().tag() ));

					if ((int)theERBArray[erb_set_index].size() > theERBAssoc) {
						// Need to evict oldest ERB entry
						rt_entry = &(theERBArray[erb_set_index].back());
						uint64_t offset_skew = (rt_entry->tag() >> (theSkewShift + theBlockSetShift)) & theOffsetMask;
						if (!theSkewOffset) {
							offset_skew = 0;
						}
						for(int32_t i = 0; i < theBlocksPerRegion; i++) {
							PhysicalMemoryAddress v_addr(rt_entry->tag() | ((i ^ offset_skew) << theBlockSetShift));
DBG_(Trace, ( << "Trying to evict ERB block " << std::hex << v_addr ));
							if ((*rt_entry)[i] >= 0) {
								int32_t way = (*rt_entry)[i];
								int32_t block_set = get_block_set(v_addr);
								RTBlockEntry *block = &(theBlockArray[block_set][way]);
								DBG_Assert(block->tag() == v_addr, ( << "Block tag " << std::hex << block->tag() << " != expected tag " << v_addr ));
								if (block->state() != ZeroSharers) {
									actions.push_back(boost::lambda::bind(theInvalidateAction, v_addr, block->sharers()));
								}

DBG_(Trace, ( << "Evicting block " << std::hex << theBlockArray[block_set][way].tag() ));
								// Mark Block not valid
								theBlockNotValidBits[block_set][way] = true;
							}
						}
DBG_(Trace, ( << "Evicting region from ERB " << std::hex << theERBArray[erb_set_index].back().tag() ));
						theERBArray[erb_set_index].pop_back();
					}
				}

				// Need to initialize the entry at 'pos'
				rt_entry = &(theRegionArray[region_set_index][pos]);
				rt_entry->reset(PhysicalMemoryAddress(region_tag));
				updateRegionLRU(region_set_index, pos);
				theRegionNotValidBits[region_set_index][pos] = false;
DBG_(Trace, ( << "Allocing region " << std::hex << rt_entry->tag() ));
			} else {
				rt_entry = &(*erb_iter);
			}
		} else {
			rt_entry = &(theRegionArray[region_set_index][region_way]);
			if (update_lru) {
				updateRegionLRU(region_set_index, region_way);
			}
		}

		DBG_Assert(rt_entry != NULL);

		// Okay, we have a region entry, now we need to find a block entry

		int32_t offset = get_offset(addr);
		int32_t way = (*rt_entry)[offset];
		int32_t block_set = get_block_set(addr);
		if (way < 0) {
			// The block we're looking for isn't allocated in the directory. we need to fix this
			// First look for invalid blocks, then use PLRU
			boost::dynamic_bitset<uint16_t>::size_type pos = theBlockNotValidBits[block_set].find_first();
			if (pos == boost::dynamic_bitset<uint16_t>::npos) {
				// No invalid block entries
				if (theBlockReplPolicy == ePLRUPolicy) {
					pos = theBlockPLRUBits[block_set].find_first();
				} else { 
					pos = selectMinBlock(block_set);
				}
				DBG_Assert(pos != boost::dynamic_bitset<uint16_t>::npos);

				RTBlockEntry *block = &(theBlockArray[block_set][pos]);
DBG_(Trace, ( << "Evicting Valid block " << std::hex << block->tag() ));
				// Need to evict the block 
				if (block->state() != ZeroSharers) {
					actions.push_back(boost::lambda::bind(theInvalidateAction, PhysicalMemoryAddress(block->tag()), block->sharers()));
				}
				// Which region does this block belong to?
				int32_t v_offset = get_offset(block->tag());
				int64_t v_r_tag(get_region_tag(block->tag()));
				int32_t v_r_set = get_region_set(block->tag());
				int32_t v_r_way = 0;
DBG_(Trace, ( << "Searching for Region containgin victim block " << std::hex << v_r_tag ));
				for(; v_r_way < theRegionAssoc; v_r_way++) {
					if (theRegionArray[v_r_set][v_r_way].tag() == v_r_tag) {
						theRegionArray[v_r_set][v_r_way][v_offset] = -1;
						break;
					}
				}
				if (v_r_way == theRegionAssoc) {
					// Look in ERB for region
					int32_t v_erb_set = get_erb_set(block->tag());
					std::list<RTDirectoryEntry>::iterator iter = theERBArray[v_erb_set].begin();
					std::list<RTDirectoryEntry>::iterator end = theERBArray[v_erb_set].end();
					for(; iter != end; iter++) {
						if (iter->tag() == v_r_tag) {
							break;
						}
					}
					DBG_Assert(iter != end, ( << "Did not find region entry for victim block!?!" ));
					(*iter)[v_offset] = -1;
				}
			}
			way = pos;
			(*rt_entry)[offset] = way;

			// Initialize block
			theBlockArray[block_set][way].reset(get_block_tag(addr));
			theBlockNotValidBits[block_set][way] = false;

DBG_(Trace, ( << "Allocating Valid block " << std::hex << theBlockArray[block_set][way].tag() ));
		}

		if (update_lru) {
			updateBlockLRU(block_set, way);
		}
		return &(theBlockArray[block_set][way]);
	}

protected:

	virtual void processRequestResponse(int32_t index, const MMType &request, MMType &response, 
										AbstractEntry_p dir_entry, PhysicalMemoryAddress address, bool off_chip) {
		// First, do default behaviour
		AbstractDirectory::processRequestResponse(index, request, response, dir_entry, address, off_chip);

		//RTBlockEntry *my_entry = dynamic_cast<RTBlockEntry*>(dir_entry);
		RTBlockEntry *my_entry(&(dynamic_cast<BlockEntryWrapper*>(dir_entry.get())->block));
		if (my_entry == NULL) {
			return;
		}

		if (my_entry->state() != ZeroSharers) {
			return;
		}

		// The block has no sharers, remove it from the directory to improve the replacement policy

		// Find the region directory entry
		RTDirectoryEntry *rt_entry = findEntry(get_region_tag(address));
		DBG_Assert(rt_entry != NULL);

		
		int32_t offset = get_offset(address);
		int32_t way = (*rt_entry)[offset];
		int32_t block_set = get_block_set(address);

		// Invalidate the block in the directory
		(*rt_entry)[offset] = -1;
		theBlockNotValidBits[block_set][way] = true;

		// If the region is not owned and has no blocks present, mark the region as invalid
		if (rt_entry->owner >= 0) {
			return;
		}

		for(int32_t i = 0; i < theBlocksPerRegion; i++) {
			if ((*rt_entry)[i] >= 0) {
				// If we find any allocated blocks, we're done
				return;
			}
		}

		// No allocated blocks in the region, mark the region invalid
		removeRegion(rt_entry);
	}

public:
	virtual void updateLRU(int32_t index, AbstractEntry_p dir_entry, PhysicalMemoryAddress address) {
		// TODO: Update Block and Region LRU HERE instead of elsewhere
	}

    virtual boost::tuple<SharingVector, SharingState, AbstractEntry_p>
            lookup(int32_t index, PhysicalMemoryAddress address, MMType req_type, std::list<boost::function<void(void)> > &xtra_actions) {


		BlockEntryWrapper_p block(new BlockEntryWrapper(*findOrCreateEntry(address, !MemoryMessage::isEvictType(req_type), xtra_actions)));
		return boost::tie((*block)->sharers(), (*block)->state(), block);
	}

    virtual boost::tuple<SharingVector, SharingState, AbstractEntry_p, bool>
            snoopLookup(int32_t index, PhysicalMemoryAddress address, MMType req_type) {

		RTBlockEntry* entry = findBlockEntry(address);
		bool valid = true;
		if (entry == NULL) {
			SharingVector sharers;
			SharingState state = ZeroSharers;
			BlockEntryWrapper_p wrapper;
			valid = false;
			return boost::tie(sharers, state, wrapper, valid);
		}
		BlockEntryWrapper_p block(new BlockEntryWrapper(*entry));
		return boost::tie((*block)->sharers(), (*block)->state(), block, valid);
	}


	void saveState( std::ostream & s, const std::string &aDirName ) {
		boost::archive::binary_oarchive oa(s);

		uint64_t set_count = theRegionSets;
		uint32_t associativity = theRegionAssoc;

		oa << set_count;
		oa << associativity;

		// Save RegionArray, ERB, and Block Array
		// Need to do this set-by-set so we can load it in set-by-set in the timing model
		for(int32_t set = 0; set < theRegionSets; set++) {
			oa & theRegionArray[set];

			std::vector<uint16_t> blocks(theRegionPLRUBits[set].num_blocks());
			to_block_range(theRegionPLRUBits[set], blocks.begin());
			oa & blocks;

			std::vector<uint16_t> not_valid_bits(theRegionNotValidBits[set].num_blocks());
			to_block_range(theRegionNotValidBits[set], not_valid_bits.begin());
			oa & not_valid_bits;
		}

		set_count = theERBSets;
		associativity = theERBAssoc;
		oa << set_count;
		oa << associativity;

		for(int32_t set = 0; set < theERBSets; set++) {
			oa & theERBArray[set];
		}

		set_count = theBlockSets;
		associativity = theBlockAssoc;

		oa << set_count;
		oa << associativity;

		for(int32_t set = 0; set < theBlockSets; set++) {
			for(int32_t way = 0; way < theBlockAssoc; way++) {
				StdDirEntrySerializer serializer(theBlockArray[set][way].getSerializer());
				oa << serializer;
			}

			std::vector<uint16_t> blocks(theBlockPLRUBits[set].num_blocks());
			to_block_range(theBlockPLRUBits[set], blocks.begin());
			oa & blocks;

			std::vector<uint16_t> not_valid_bits(theBlockNotValidBits[set].num_blocks());
			to_block_range(theBlockNotValidBits[set], not_valid_bits.begin());
			oa & not_valid_bits;
		}
	}

	bool loadState( std::istream & s, const std::string &aDirName ) {
		boost::archive::binary_iarchive ia(s);

		uint64_t set_count = 0;
		uint32_t associativity = 0;

		ia >> set_count;
		ia >> associativity;

		DBG_Assert( set_count == (uint64_t)theRegionSets, ( << "Error loading directory state. Flexpoint contains " << set_count << " region sets but simulator configured for " << theRegionSets << " region sets." ));
		DBG_Assert( associativity == (uint64_t)theRegionAssoc, ( << "Error loading directory state. Flexpoint contains " << associativity << "-way region sets but simulator configured for " << theRegionAssoc << "-way region sets." ));

		for(int32_t set = 0; set < theRegionSets; set++) {
			ia >> theRegionArray[set];

			std::vector<uint16_t> blocks(theRegionPLRUBits[set].num_blocks());
			ia & blocks;
			from_block_range(blocks.begin(), blocks.end(), theRegionPLRUBits[set]);

			std::vector<uint16_t> not_valid_bits(theRegionPLRUBits[set].num_blocks());
			ia & not_valid_bits;
			from_block_range(not_valid_bits.begin(), not_valid_bits.end(), theRegionNotValidBits[set]);
		}

		ia >> set_count;
		ia >> associativity;
		DBG_Assert( set_count == (uint64_t)theERBSets, ( << "Error loading directory state. Flexpoint contains " << set_count << " erb sets but simulator configured for " << theERBSets << " erb sets." ));
		DBG_Assert( associativity == (uint64_t)theERBAssoc, ( << "Error loading directory state. Flexpoint contains " << associativity << "-way ERB sets but simulator configured for " << theERBAssoc << "-way ERB sets." ));
		for(int32_t set = 0; set < theERBSets; set++) {
			ia >> theERBArray[set];
		}

		ia >> set_count;
		ia >> associativity;

		DBG_Assert( set_count == (uint64_t)theBlockSets, ( << "Error loading directory state. Flexpoint contains " << set_count << " sets but simulator configured for " << theBlockSets << " sets." ));
		DBG_Assert( associativity == (uint64_t)theBlockAssoc, ( << "Error loading directory state. Flexpoint contains " << associativity << "-way sets but simulator configured for " << theBlockAssoc << "-way sets." ));

		for(int32_t set = 0; set < theBlockSets; set++) {
			for(int32_t way = 0; way < theBlockAssoc; way++) {
				StdDirEntrySerializer serializer;
				ia >> serializer;
				theBlockArray[set][way] = serializer;
			}

			std::vector<uint16_t> blocks(theBlockPLRUBits[set].num_blocks());
			ia & blocks;
			from_block_range(blocks.begin(), blocks.end(), theBlockPLRUBits[set]);

			std::vector<uint16_t> not_valid_bits(theBlockPLRUBits[set].num_blocks());
			ia & not_valid_bits;
			from_block_range(not_valid_bits.begin(), not_valid_bits.end(), theBlockNotValidBits[set]);
		}
		return true;
	}

	static AbstractDirectory* createInstance(std::list<std::pair<std::string, std::string> > &args) {
		RTDirectory* directory = new RTDirectory();

		directory->parseGenericOptions(args);

		std::list<std::pair<std::string,std::string> >::iterator iter = args.begin();
		for(; iter != args.end(); iter++) {
			if (strcasecmp(iter->first.c_str(), "sets") == 0) {
				directory->theBlockSets = boost::lexical_cast<int>(iter->second);
			} else if (strcasecmp(iter->first.c_str(), "assoc") == 0) {
				directory->theBlockAssoc = boost::lexical_cast<int>(iter->second);
			} else if (strcasecmp(iter->first.c_str(), "region_sets") == 0) {
				directory->theRegionSets = boost::lexical_cast<int>(iter->second);
			} else if (strcasecmp(iter->first.c_str(), "region_assoc") == 0) {
				directory->theRegionAssoc = boost::lexical_cast<int>(iter->second);
			} else if (strcasecmp(iter->first.c_str(), "erb_sets") == 0) {
				directory->theERBSets = boost::lexical_cast<int>(iter->second);
			} else if (strcasecmp(iter->first.c_str(), "erb_assoc") == 0) {
				directory->theERBAssoc = boost::lexical_cast<int>(iter->second);
			} else if (strcasecmp(iter->first.c_str(), "rsize") == 0) {
				directory->theRegionSize = boost::lexical_cast<int>(iter->second);
            } else if (strcasecmp(iter->first.c_str(), "skew") == 0) {
				directory->theSkewSet = boost::lexical_cast<bool>(iter->second);
            } else if (strcasecmp(iter->first.c_str(), "skew_offset") == 0) {
				directory->theSkewOffset = boost::lexical_cast<bool>(iter->second);
            } else if (strcasecmp(iter->first.c_str(), "use_region_plru") == 0) {
				bool use_plru = boost::lexical_cast<bool>(iter->second);
				if (use_plru) {
					directory->theRegionReplPolicy = ePLRUPolicy;
				} else {
					directory->theRegionReplPolicy = eMinPolicy;
				}
            } else if (strcasecmp(iter->first.c_str(), "use_block_plru") == 0) {
				bool use_plru = boost::lexical_cast<bool>(iter->second);
				if (use_plru) {
					directory->theBlockReplPolicy = ePLRUPolicy;
				} else {
					directory->theBlockReplPolicy = eMinPolicy;
				}
			}
		}

		// directory->initialize();
		// Initialization called after other parameters have been set

		return directory;

	}

	static const std::string name;

};

REGISTER_DIRECTORY_TYPE(RTDirectory, "RegionTracker");

}; // namespace
