// DO-NOT-REMOVE begin-copyright-block 
//
// Redistributions of any form whatsoever must retain and/or include the
// following acknowledgment, notices and disclaimer:
//
// This product includes software developed by Carnegie Mellon University.
//
// Copyright 2010 by Mohammad Alisafaee, Eric Chung, Michael Ferdman, Brian Gold,
// Jangwoo Kim, Pejman Lotfi-Kamran, Onur Kocberber, Ippokratis Pandis, 
// Djordje Jevdjic, Jared Smolens, Stephen Somogyi, Evangelos Vlachos, 
// Stavros Volos, Jason Zebchuk, Babak Falsafi, Nikos Hardavellas and Tom Wenisch 
// for the SimFlex Project, Computer Architecture Lab at Carnegie Mellon,
// Carnegie Mellon University.
//
// For more information, see the SimFlex project website at:
//   http://www.ece.cmu.edu/~simflex
//
// You may not use the name "Carnegie Mellon University" or derivations
// thereof to endorse or promote products derived from this software.
//
// If you modify the software you must place a notice on or within any
// modified version provided or made available to any third party stating
// that you have modified the software.  The notice shall include at least
// your name, address, phone number, email address and the date and purpose
// of the modification.
//
// THE SOFTWARE IS PROVIDED "AS-IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER
// EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT LIMITED TO ANY WARRANTY
// THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS OR BE ERROR-FREE AND ANY
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
// TITLE, OR NON-INFRINGEMENT.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
// BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO DIRECT, INDIRECT,
// SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM, OR IN
// ANY WAY CONNECTED WITH THIS SOFTWARE (WHETHER OR NOT BASED UPON WARRANTY,
// CONTRACT, TORT OR OTHERWISE).
//
// DO-NOT-REMOVE end-copyright-block   

#include <cmath>

#include <boost/regex.hpp>
#include <boost/throw_exception.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/function.hpp>
#include <boost/lambda/bind.hpp>
namespace ll = boost::lambda;

#include <core/target.hpp>
#include <core/types.hpp>
#include <core/simics/api_wrappers.hpp>

#include "mai_api.hpp"



namespace Flexus {
namespace Simics {

void onInterrupt (void * aPtr, Simics::API::conf_object_t * anObj, long long aVector);

struct InterruptManager {
  int theHandle;
  typedef std::map< Simics::API::conf_object_t *, boost::function< void( long long) > > dispatch_map;
  dispatch_map theInterruptDispatch;

  void registerHAP() {
		theHandle = API::SIM_hap_add_callback( "Core_Asynchronous_Trap", reinterpret_cast< API::obj_hap_func_t>(&onInterrupt), 0);
  }

  void registerCPU(API::conf_object_t * aCPU, boost::function< void( long long) > aDispatchFn ) {
    if (theInterruptDispatch.empty()) {
      registerHAP();
    }
    theInterruptDispatch.insert( std::make_pair( aCPU, aDispatchFn) );
  }

  void interrupt(API::conf_object_t * anObj, long long aVector ) {
    dispatch_map::iterator iter = theInterruptDispatch.find(anObj);
    if (iter != theInterruptDispatch.end()) {
      iter->second( aVector );
    }
  }

} theInterruptManager;

void onInterrupt (void * aPtr, API::conf_object_t * anObj, long long aVector) {
  theInterruptManager.interrupt( anObj, aVector );
}

#if FLEXUS_TARGET_IS(v9)

void v9ProcessorImpl::initialize() {
  DBG_( Dev, ( << "CPU[" << Simics::APIFwd::SIM_get_processor_number(*this) << "] Registering for interrupts "));
  theInterruptManager.registerCPU( *this, ll::bind( &v9ProcessorImpl::handleInterrupt, this, ll::_1 ) );
}


void v9ProcessorImpl::handleInterrupt( long long aVector) {
  DBG_( Dev, ( << "CPU[" << Simics::APIFwd::SIM_get_processor_number(*this) << "] Interrupt " << std::hex << aVector << std::dec ));
  thePendingInterrupt = aVector;
}


static const int kTL = 46;
static const int kTT1 = 78;

int v9ProcessorImpl::getPendingException() const {
  int tl = readRegister( kTL );
  if (tl > 0) {
    return readRegister( kTT1 + tl - 1);
  } else {
    return 0;
  }
}

int v9ProcessorImpl::getPendingInterrupt() const {
  if (thePendingInterrupt == API::Sim_PE_No_Exception) {
    return 0;
  }
  return thePendingInterrupt;
}


int v9ProcessorImpl::advance(bool anAcceptInterrupt) {
  int exception = 0;

  if (! theInterruptsConnected) {
    theInterruptsConnected = true;
    initialize();
  }

  if (thePendingInterrupt != API::Sim_PE_No_Exception && anAcceptInterrupt) {
    //Try to take the interrupt now.
    int interrupt = thePendingInterrupt;
    switch(API::SIM_instruction_handle_interrupt(*this, thePendingInterrupt)) {
      case API::Sim_IE_No_Exception:
        thePendingInterrupt = API::Sim_PE_No_Exception;
        DBG_( Crit, ( << "Got Sim_IE_No_Exception trying to take an interrupt" ) );
        break;
      case API::Sim_IE_OK:
        thePendingInterrupt = API::Sim_PE_No_Exception;
        return interrupt;
      case API::Sim_IE_Interrupts_Disabled:
        // OK try later */
        DBG_( Crit, ( << "Got Sim_IE_Interruts_Disabled trying to take an interrupt on CPU[" << Simics::APIFwd::SIM_get_processor_number(*this) << "]"  ) );
        return -1;
      case API::Sim_IE_Illegal_Interrupt_Point:
        // OK try later */
        DBG_( Crit, ( << "Got Sim_IE_Illegal_Interrupt_Point trying to take an interrupt" ) );
        return -1;
      case API::Sim_IE_Step_Breakpoint:
        //Return to prompt next cycle
        API::SIM_break_cycle(*this, 0);
        break;
      default:
        DBG_( Crit, ( << "Bad return value from SIM_instruction_handle_interrupt" ) );
    }
  }

  int except = API::SIM_get_pending_exception();
  if (except != API::SimExc_No_Exception) {
        DBG_( Crit, ( << "Pending Exception: " << except ) );
   }

  //Normal instruction processing
    API::instruction_id_t inst = API::SIM_instruction_begin(*this);
    API::SIM_instruction_insert(0, inst);

    bool retry= false;
    do {
      retry = false;
      API::instruction_error_t err = API::SIM_instruction_commit(inst);
      if (err == API::Sim_IE_Exception ) {
        err = API::SIM_instruction_handle_exception(inst);
        exception = getPendingException();
        DBG_( Verb, ( << "Exception raised by CPU[" << Simics::APIFwd::SIM_get_processor_number(*this) << "]: " << err << " Exception #:" << exception) );
        API::SIM_instruction_end(inst);
        if (err != API::Sim_IE_OK) {
          DBG_( Crit, ( << "Got an error trying to handle exception on CPU[" << Simics::APIFwd::SIM_get_processor_number(*this) << ": " << err << " exception was: " << exception) );
        }
        return exception; //Instruction is already ended
      } else if (err == API::Sim_IE_Code_Breakpoint || err == API::Sim_IE_Step_Breakpoint || err == API::Sim_IE_Hap_Breakpoint ) {
        DBG_( Verb, ( << "Triggerred Breakpoint on CPU[" << Simics::APIFwd::SIM_get_processor_number(*this) << "]" ) );
        API::SIM_break_cycle(*this, 0);
        retry = true;
      } else if (err != API::Sim_IE_OK) {
        DBG_( Crit, ( << "Got an error trying to advance CPU[" << Simics::APIFwd::SIM_get_processor_number(*this) << "]: " << err) );
        //Some error other than an exception
      }
    } while( retry );
    API::SIM_instruction_end(inst);

  return exception;
}

unsigned long long endianFlip(unsigned long long val, int aSize);

unsigned long long v9ProcessorImpl::interruptRead(VirtualMemoryAddress anAddress, int anASI) const {
  DBG_( Verb, ( << "Reading " << anAddress << " in ASI: " << std::hex << anASI << std::dec ) );
  API::logical_address_t addr(anAddress);
  unsigned long long aValue = 0;
  API::v9_memory_transaction_t xact;
  memset( &xact, 0, sizeof(API::v9_memory_transaction_t ) );
  xact.priv = 1;
  #if SIM_VERSION < 1200
    xact.align_kind = API::Align_Natural;
  #else
    //align_kind was replaced by access_type in Simics 2.2.x
    xact.access_type = API::V9_Access_Normal;
  #endif
  xact.address_space = anASI;
  xact.s.logical_address = addr;
  xact.s.physical_address = addr;
  xact.s.size = 8;
  xact.s.type = API::Sim_Trans_Load;
  xact.s.inquiry = 1;
  xact.s.ini_type = API::Sim_Initiator_CPU_UIII;
  xact.s.ini_ptr = theProcessor;
  xact.s.real_address = reinterpret_cast<char*>(&aValue);
  xact.s.exception = API::Sim_PE_No_Exception;

  API::exception_type_t except = sparc()->access_asi_handler(theProcessor, &xact);
  if (except != API::Sim_PE_No_Exception) {
    DBG_( Dev, ( << "except: " << except  ) );
  }
  return endianFlip(aValue, 8);
}


#endif //FLEXUS_TARGET_IS(v9)

ProcessorMapper* ProcessorMapper::theMapper = NULL;

struct cpu_desc_t {
	int vm;
	int cpu;
	int simics_id;
	bool is_client;
	cpu_desc_t(int v, int c, int s, bool client) : vm(v), cpu(c), simics_id(s), is_client(client) {}
};

struct cpu_desc_sorter_t {
	bool operator()(const cpu_desc_t &a, const cpu_desc_t &b) const {
		if (a.vm == b.vm) {
			if (a.cpu == b.cpu) {
				return (b.is_client);
			} else {
				return (a.cpu < b.cpu);
			}
		} else {
			return (a.vm < b.vm);
		}
	}
};

ProcessorMapper::ProcessorMapper() {
	// Find the Processors, Find All of them

	// Clear all exceptions so that get_all_processors works properly.

	API::SIM_clear_exception();

	API::attr_value_t proc_list = API::SIM_get_all_processors();
	DBG_Assert( proc_list.kind == API::Sim_Val_List, ( << "SIM_get_all_processors() did NOT return a LIST" ));
	int proc_count = proc_list.u.list.size;
	std::vector<cpu_desc_t> cpu_list;
	int num_flexus_cpus = 0;
	int num_client_cpus = 0;
	int max_vm = -1;

	boost::regex vm_expression("machine([0-9]*)_((?:)|(?:server_)|(?:client_))cpu([0-9]*)");
	boost::regex non_vm_expression("((?:)|(?:server_)|(?:client_))cpu([0-9]*)");
	DBG_(Crit, ( << "Searching " << proc_count << " cpus." ));
	for(int i = 0; i < proc_count; i++) {
		API::conf_object_t *cpu = proc_list.u.list.vector[i].u.object;
		int simics_id = APIFwd::SIM_get_processor_number(cpu);
		DBG_(Crit, ( << "Processor " << i << ": " << cpu->name << " - CPU " << simics_id ));

		boost::cmatch what;
		if (boost::regex_match(cpu->name, what, vm_expression)) {
			int vm = boost::lexical_cast<int>(what[1]);
			int cpu = boost::lexical_cast<int>(what[3]);
			bool is_client = false;
			DBG_Assert(max_vm >= 0 || (num_flexus_cpus == 0 && num_client_cpus == 0), ( << "Combined VM/Non-VM workload detected: max_vm = " << max_vm << ", flexus = " << num_flexus_cpus << ", clients = " << num_client_cpus << ", vm = " << vm ));
			if (what[2] == "client_") {
				is_client = true;
				num_client_cpus++;
			} else {
				num_flexus_cpus++;
			}
			if (vm > max_vm) {
				max_vm = vm;
			}
			cpu_list.push_back(cpu_desc_t(vm, cpu, simics_id, is_client));
		} else if (boost::regex_match(cpu->name, what, non_vm_expression)) {
DBG_(Crit, ( << "Found CPU: '" << what[1] << "' - " << what[2] ));
			int vm = 0;
			int cpu = boost::lexical_cast<int>(what[2]);
			bool is_client = false;
			if (what[1] == "client_") {
				is_client = true;
				num_client_cpus++;
			} else {
				num_flexus_cpus++;
			}
			DBG_Assert(max_vm == -1);
			cpu_list.push_back(cpu_desc_t(vm, cpu, simics_id, is_client));
		}
	}
	// Don't forget to free the processor list
	SIM_free_attribute(proc_list);

	DBG_(Crit, ( << "Found " << num_flexus_cpus << " Flexus CPUs and " << num_client_cpus << " Client CPUs in " << (max_vm + 1) << " VMs"));
	theProcMap.resize(num_flexus_cpus, std::make_pair(0,0));
	theClientMap.resize(num_client_cpus, 0);
	theReverseMap.resize(proc_count, std::make_pair(0, false));

	int client_index = 0;
	int flexus_index = 0;

	std::sort(cpu_list.begin(), cpu_list.end(), cpu_desc_sorter_t());

	std::vector<cpu_desc_t>::iterator iter = cpu_list.begin();
#if 0
	for(; iter != cpu_list.end(); iter++) {
		theReverseMap[iter->simics_id] = std::make_pair(flexus_index, iter->is_client);
		if (iter->is_client) {
			theClientMap[client_index] = iter->simics_id;
			DBG_(Verb, ( << "Client[" << client_index << "] = " << iter->simics_id ));
			client_index++;
		} else {
			theProcMap[flexus_index].first = iter->simics_id;
			theProcMap[flexus_index].second = iter->vm;
			DBG_(Verb, ( << "theProcMap[" << flexus_index << "] = (" << iter->simics_id << ", " << iter->vm << ")" ));
			flexus_index++;
		}
	}
#endif
	theNumVMs = max_vm + 1;

	if (theNumVMs == 0) {
		theNumVMs = 1;
	}


	// Assume that all the vms are the same size, and square
	int vms_per_row = (int)std::ceil(std::sqrt(theNumVMs));
	int cores_per_vm = num_flexus_cpus / theNumVMs;
    int num_vm_rows = (int)std::ceil(std::sqrt(cores_per_vm));
    int num_rows = num_vm_rows * vms_per_row;

	DBG_(Crit, ( << "VMS per row = " << vms_per_row << ", CPVM = " << cores_per_vm << ", NVMR = " << num_vm_rows << ", NumRow = " << num_rows ));

	for(; iter != cpu_list.end(); iter++) {
		theReverseMap[iter->simics_id] = std::make_pair(flexus_index, iter->is_client);
		if (iter->is_client) {
			theReverseMap[iter->simics_id] = std::make_pair(-1, iter->is_client);
			theClientMap[client_index] = iter->simics_id;
			DBG_(Crit, ( << "Client[" << client_index << "] = " << iter->simics_id ));
			client_index++;
		} else {

			// re-map flexus index, so VMs are organized in a nice grid
			int vm = flexus_index / cores_per_vm;
            int core = flexus_index % cores_per_vm;
            int row_in_vm = core / num_vm_rows;
            int col_in_vm = core % num_vm_rows;
            int row_of_vm = vm / vms_per_row;
            int col_of_vm = vm % vms_per_row;

            int abs_row = row_in_vm + row_of_vm*num_vm_rows;
            int abs_col = col_in_vm + col_of_vm*num_vm_rows;
    
            int new_index = abs_row*num_rows + abs_col;

			theReverseMap[iter->simics_id] = std::make_pair(new_index, iter->is_client);
			theProcMap[new_index].first = iter->simics_id;
			theProcMap[new_index].second = iter->vm;
			DBG_(Crit, ( << "theProcMap[" << new_index << "] = (" << iter->simics_id << ", " << iter->vm << ") (abs_index = " << flexus_index << ")" ));
			flexus_index++;
		}
	}

	DBG_(Dev, ( << "Finished creating Processor Mapper." ));
	API::SIM_clear_exception();
}

int ProcessorMapper::mapFlexusIndex2ProcNum(int index) {
	if (theMapper == NULL) {
		theMapper = new ProcessorMapper();
	}
	return theMapper->theProcMap[index].first;
}

int ProcessorMapper::mapClientNum2ProcNum(int index) {
	if (theMapper == NULL) {
		theMapper = new ProcessorMapper();
	}
	return theMapper->theClientMap[index];
}

int ProcessorMapper::mapProcNum2FlexusIndex(int index) {
	if (theMapper == NULL) {
		theMapper = new ProcessorMapper();
	}
	return theMapper->theReverseMap[index].first;
}

int ProcessorMapper::mapFlexusIndex2VM(int index) {
	if (theMapper == NULL) {
		theMapper = new ProcessorMapper();
	}
	if (theMapper->theNumVMs == 0) {
		return 0;
	}
	return theMapper->theProcMap[index].second;
}

int ProcessorMapper::numVMs() {
	if (theMapper == NULL) {
		theMapper = new ProcessorMapper();
		DBG_Assert(theMapper != NULL);
	}
	return theMapper->theNumVMs;
}

int ProcessorMapper::numClients() {
	if (theMapper == NULL) {
		theMapper = new ProcessorMapper();
	}
	return (int)(theMapper->theClientMap.size());
}

int ProcessorMapper::numProcessors() {
	if (theMapper == NULL) {
		theMapper = new ProcessorMapper();
	}
	return (int)(theMapper->theProcMap.size());
}

} //end Namespace Simics
} //end namespace Flexus

